
//	HID Minidriver for Microsoft Xbox 360 Controller

#include	"X360C.h"

//----------------------------------------------------------------------------------------------
//	CompleteRequest
//----------------------------------------------------------------------------------------------
VOID CompleteRequest(
	 IN PIRP		Irp
	,IN NTSTATUS	Status
	,IN ULONG		Information )
{
	//	I/O v
	Irp->IoStatus.Status		= Status;
	Irp->IoStatus.Information	= Information;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
}

//----------------------------------------------------------------------------------------------
//	SendRequest
//----------------------------------------------------------------------------------------------
NTSTATUS SendRequest(
	 IN PDEVICE_OBJECT	DeviceObject
	,IN PIRP			Irp )
{
	//	ʃhCo I/O v𔭍s
	IoSkipCurrentIrpStackLocation( Irp );
	return( IoCallDriver( GET_NEXT_DEVICE_OBJECT( DeviceObject ), Irp ) );
}

//----------------------------------------------------------------------------------------------
//	SendAndWaitRequest
//----------------------------------------------------------------------------------------------
NTSTATUS SendAndWaitRequest(
	 IN PDEVICE_OBJECT	DeviceObject
	,IN PIRP			Irp )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;
	KEVENT		WaitEvent;

	//	ʃhCo I/O v𔭍s
	KeInitializeEvent( &WaitEvent, NotificationEvent, FALSE );
	IoCopyCurrentIrpStackLocationToNext( Irp );
	IoSetCompletionRoutine(
		 Irp
		,(PIO_COMPLETION_ROUTINE)IoCompletionRoutine
		,(PVOID)&WaitEvent
		,TRUE
		,TRUE
		,TRUE );
	Status	= IoCallDriver( GET_NEXT_DEVICE_OBJECT( DeviceObject ), Irp );

	//	ʃhCȍ܂őҋ@
	if( Status == STATUS_PENDING )
	{
		KeWaitForSingleObject( &WaitEvent, Executive, KernelMode, FALSE, NULL );
		Status	= Irp->IoStatus.Status;
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	IoCompletionRoutine
//----------------------------------------------------------------------------------------------
NTSTATUS IoCompletionRoutine(
	 IN PDEVICE_OBJECT	DeviceObject
	,IN PIRP			Irp
	,IN PKEVENT			Event )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_MORE_PROCESSING_REQUIRED;

	//	I/O v̊ʒm
	KeSetEvent( Event, 0, FALSE );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	IncrementReference
//----------------------------------------------------------------------------------------------
NTSTATUS IncrementReference( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;

	//	QƃJE^Z
	InterlockedIncrement( &DeviceExtension->ReferenceCount );

	//	hCȍԂ͗LEH
	if( DeviceExtension->DriverState == Disable )
	{
		if( InterlockedDecrement( &DeviceExtension->ReferenceCount ) == 0 )
		{
			//	QƏICxg𑗂
			KeSetEvent( &DeviceExtension->ReferenceEnd, 0, FALSE );
			Status	= STATUS_DELETE_PENDING;
		}
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	DecrementReference
//----------------------------------------------------------------------------------------------
NTSTATUS DecrementReference( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;

	//	QƃJE^Z
	if( InterlockedDecrement( &DeviceExtension->ReferenceCount ) == 0 )
	{
		//	QƏICxg𑗂
		KeSetEvent( &DeviceExtension->ReferenceEnd, 0, FALSE );
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	GetRegistryLong
//----------------------------------------------------------------------------------------------
ULONG GetRegistryLong(
	 IN PWSTR	SubKey
	,IN PWSTR	Name
	,IN ULONG	DefaultValue )
{
	//	ϐ錾
	NTSTATUS					Status		= STATUS_SUCCESS;
	ULONG						DefaultData	= DefaultValue;
	ULONG						LongValue;
	RTL_QUERY_REGISTRY_TABLE	QueryTable[3];

	//	₢킹e[uݒ肷
	RtlZeroMemory( QueryTable, sizeof( QueryTable ) );
	QueryTable[0].Flags			= RTL_QUERY_REGISTRY_SUBKEY;
	QueryTable[0].Name			= SubKey;
	QueryTable[0].EntryContext	= NULL;
	QueryTable[1].Flags			= RTL_QUERY_REGISTRY_DIRECT;
	QueryTable[1].Name			= Name;
	QueryTable[1].EntryContext	= &LongValue;
	QueryTable[1].DefaultType	= REG_DWORD;
	QueryTable[1].DefaultData	= &DefaultData;
	QueryTable[1].DefaultLength	= sizeof( DefaultData );

	//	WXgɖ₢킹s
	Status	= RtlQueryRegistryValues(
				 RTL_REGISTRY_SERVICES
				,NULL
				,QueryTable
				,NULL
				,NULL );
	if( NT_SUCCESS( Status ) )
	{
		return( LongValue );
	} else {
		return( DefaultData );
	}
}

//----------------------------------------------------------------------------------------------
//	GetRegistryBinary
//----------------------------------------------------------------------------------------------
ULONG GetRegistryBinary(
	 IN PWSTR	Path
	,IN PWSTR	Value
	,IN CHAR *	Buf
	,IN LONG	BufSize )
{
	//	ϐ錾
	NTSTATUS						Status		= STATUS_SUCCESS;
	UNICODE_STRING					RegistryPath;
	UNICODE_STRING					RegistryValue;
	OBJECT_ATTRIBUTES				ObjAttribute;
	HANDLE							Key;
	LONG							ValueSize	= 0;
	PKEY_VALUE_PARTIAL_INFORMATION	ValueInfo;

	//	WXg pX쐬
	RtlInitUnicodeString( &RegistryPath, Path );
	//	IuWFNg쐬
	InitializeObjectAttributes(
		 &ObjAttribute
		,&RegistryPath
		,OBJ_CASE_INSENSITIVE
		,NULL
		,NULL );

	//	WXgJ
	Status	= ZwOpenKey( &Key, KEY_QUERY_VALUE, &ObjAttribute );
	if( NT_SUCCESS( Status ) )
	{
		//	WXgl쐬
		RtlInitUnicodeString( &RegistryValue, Value );

		//	WXgl̃TCY擾
		Status	= ZwQueryValueKey(
					 Key
					,&RegistryValue
					,KeyValuePartialInformation
					,NULL
					,0
					,&ValueSize );
		if( ( Status == STATUS_OBJECT_NAME_NOT_FOUND ) || ( ValueSize == 0 ) )
		{
			ZwClose( Key );
			return( Status );
		}

		//	WXgl擾obt@쐬
		ValueInfo	= (PKEY_VALUE_PARTIAL_INFORMATION)ExAllocatePool( NonPagedPool, ValueSize );
		if( ValueInfo == NULL )
		{
			ZwClose( Key );
			return( Status );
		}

		//	WXgl擾
		Status	= ZwQueryValueKey(
					 Key
					,&RegistryValue
					,KeyValuePartialInformation
					,ValueInfo
					,ValueSize
					,&ValueSize );
		if( !NT_SUCCESS( Status ) )
		{
			ExFreePool( ValueInfo );
			ZwClose( Key );
			return( Status );
		}

		//	擾eޔ
		RtlCopyMemory( Buf, ValueInfo->Data, BufSize );

		//	obt@J
		ExFreePool( ValueInfo );

		//	WXg
		ZwClose( Key );
	}

	return( 0 );
}
