
//	HID Minidriver for Microsoft Xbox 360 Controller

#include	"X360C.h"

//----------------------------------------------------------------------------------------------
//	IRP_MJ_PNP
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_PnP(
	 IN PDEVICE_OBJECT	DeviceObject
	,IN PIRP			Irp )
{
	//	ϐ錾
	NTSTATUS			Status			= STATUS_SUCCESS;
	PDEVICE_EXTENSION	DeviceExtension	= GET_MINIDRIVER_DEVICE_EXTENSION( DeviceObject );
	PIO_STACK_LOCATION	CurrentIrpStack;

	//	QƃJE^Z
	Status	= IncrementReference( DeviceExtension );
	if( !NT_SUCCESS( Status ) )
	{
		CompleteRequest( Irp, Status, 0 );
		return( Status );
	}

	//	}Ci[ t@NVɂďU蕪
	CurrentIrpStack	= IoGetCurrentIrpStackLocation( Irp );
	switch( CurrentIrpStack->MinorFunction )
	{
		case IRP_MN_START_DEVICE:
			Status	= X360C_StartDevice( DeviceObject, DeviceExtension, Irp );
			break;

		case IRP_MN_QUERY_CAPABILITIES:
			Status	= X360C_QueryCapabilities( DeviceObject, Irp, CurrentIrpStack );
			break;

		case IRP_MN_SURPRISE_REMOVAL:
			Status	= X360C_SurpriseRemoval( DeviceObject, DeviceExtension, Irp );
			break;

		case IRP_MN_REMOVE_DEVICE:
			Status	= X360C_RemoveDevice( DeviceObject, DeviceExtension, Irp );
			break;

		default:
			//	IRP_MJ_PNP vʃhCoɑ
			Status	= SendRequest( DeviceObject, Irp );
			break;
	}

	//	QƃJE^Z
	DecrementReference( DeviceExtension );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	IRP_MN_START_DEVICE
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_StartDevice(
	 IN PDEVICE_OBJECT		DeviceObject
	,IN PDEVICE_EXTENSION	DeviceExtension
	,IN PIRP				Irp )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;

	//	 IRP_MN_START_DEVICE vʃhCoɑ
	Status	= SendAndWaitRequest( DeviceObject, Irp );
	if( !NT_SUCCESS( Status ) )
	{
		CompleteRequest( Irp, Status, 0 );
		return( Status );
	}

	//	foCXLɂ
	Status	= X360C_EnableDevice( DeviceObject, DeviceExtension, Irp );

	//	IRP_MN_START_DEVICE v
	CompleteRequest( Irp, Status, 0 );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	IRP_MN_QUERY_CAPABILITIES
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_QueryCapabilities(
	 IN PDEVICE_OBJECT		DeviceObject
	,IN PIRP				Irp
	,IN PIO_STACK_LOCATION	CurrentIrpStack )
{
	//	ϐ錾
	NTSTATUS				Status	= STATUS_SUCCESS;
	PDEVICE_CAPABILITIES	DeviceCapabilities;

	//	 IRP_MN_QUERY_CAPABILITIES vʃhCoɑ
	Status	= SendAndWaitRequest( DeviceObject, Irp );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	ʃhCoŐݒ肳ꂽfoCX̋@\ύX
	DeviceCapabilities	= CurrentIrpStack->Parameters.DeviceCapabilities.Capabilities;
	DeviceCapabilities->SurpriseRemovalOK	= TRUE;

	//	IRP_MN_QUERY_CAPABILITIES v
	CompleteRequest( Irp, Status, 0 );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	IRP_MN_SURPRISE_REMOVAL
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_SurpriseRemoval(
	 IN PDEVICE_OBJECT		DeviceObject
	,IN PDEVICE_EXTENSION	DeviceExtension
	,IN PIRP				Irp )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;

	//	foCX𖳌ɂ
	X360C_DisableDevice( DeviceExtension );

	//	IRP_MN_SURPRISE_REMOVAL vʃhCoɑ
	Status	= SendRequest( DeviceObject, Irp );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	IRP_MN_REMOVE_DEVICE
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_RemoveDevice(
	 IN PDEVICE_OBJECT		DeviceObject
	,IN PDEVICE_EXTENSION	DeviceExtension
	,IN PIRP				Irp )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;

	//	foCX𖳌ɂ
	X360C_DisableDevice( DeviceExtension );

	//	IRP_MN_REMOVE_DEVICE vʃhCoɑ
	Status	= SendRequest( DeviceObject, Irp );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_EnableDevice
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_EnableDevice(
	 IN PDEVICE_OBJECT		DeviceObject
	,IN PDEVICE_EXTENSION	DeviceExtension
	,IN PIRP				Irp )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;
	HANDLE		ThreadHandle;

	//	USB |[g
	Status	= X360C_UsbSetupPort( DeviceExtension );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	foCXLqq擾
	Status	= X360C_UsbGetDeviceDescriptor( DeviceExtension );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	RtBM[VLqq擾
	Status	= X360C_UsbGetConfigurationDescriptor( DeviceExtension );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	C^[tFCXLqq擾
	Status	= X360C_UsbGetInterfaceDescriptor( DeviceExtension );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	RtBM[VI
	Status	= X360C_UsbSelectConfiguration( DeviceExtension );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	foCX̏ԂLɂ
	DeviceExtension->DeviceState	= Enable;

	//	[J XbhJn
	DeviceExtension->ThreadState	= Enable;
	PsCreateSystemThread(
		 &ThreadHandle
		,THREAD_ALL_ACCESS
		,NULL
		,NULL
		,NULL
		,X360C_WorkerThreadIn
		,DeviceExtension );
	ObReferenceObjectByHandle(
		 ThreadHandle
		,THREAD_ALL_ACCESS
		,NULL
		,KernelMode
		,&DeviceExtension->ThreadObjectIn
		,NULL );
	ZwClose( ThreadHandle );

	if( DeviceExtension->OutputPipeHandle != NULL )
	{
		//	[J XbhJn
		PsCreateSystemThread(
			 &ThreadHandle
			,THREAD_ALL_ACCESS
			,NULL
			,NULL
			,NULL
			,X360C_WorkerThreadOut
			,DeviceExtension );
		ObReferenceObjectByHandle(
			 ThreadHandle
			,THREAD_ALL_ACCESS
			,NULL
			,KernelMode
			,&DeviceExtension->ThreadObjectOut
			,NULL );
		ZwClose( ThreadHandle );
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_WorkerThreadIn
//----------------------------------------------------------------------------------------------
void X360C_WorkerThreadIn( IN PVOID Context )
{
	//	ϐ錾
	NTSTATUS			Status			= STATUS_SUCCESS;
	PDEVICE_EXTENSION	DeviceExtension	= Context;
	LARGE_INTEGER		TimeOut;

	//	Xbh̗D揇ʂ
	KeSetPriorityThread( KeGetCurrentThread(), LOW_REALTIME_PRIORITY );

	while(TRUE)
	{
		//	Xbh̏I`FbN
		if( DeviceExtension->ThreadState == Disable )
		{
			break;
		}

		//	WXgݒǂݍ
		if( DeviceExtension->ResetDevice == TRUE )
		{
			X360C_LoadSetting( DeviceExtension );
			DeviceExtension->ResetRingLight	= TRUE;
			DeviceExtension->ResetDevice	= FALSE;
		}

		//	|[g̓o͂s
		Status	= X360C_UsbGetReport( DeviceExtension );
		if( !NT_SUCCESS( Status ) )
		{
			//	pCvZbg
			if(	Status	!= STATUS_CANCELLED
			&&	Status	!= STATUS_DEVICE_NOT_CONNECTED )
			{
				X360C_UsbResetPipe( DeviceExtension, DeviceExtension->InputPipeHandle );
			}
			//	16 ms ҋ@
			TimeOut.QuadPart	= -16 * 1000 * 10;
			KeDelayExecutionThread( KernelMode, FALSE, &TimeOut );
		} else {
			//	ҋ@
			TimeOut.QuadPart	= INTERVAL;
			KeDelayExecutionThread( KernelMode, FALSE, &TimeOut );
		}
	}

	//	XbhJ
	ObDereferenceObject( DeviceExtension->ThreadObjectIn );
	DeviceExtension->ThreadObjectIn	= NULL;

	//	XbhIƂʒm
	KeSetEvent( &DeviceExtension->ThreadExitingIn, 0 ,FALSE );

	//	XbhI
	PsTerminateSystemThread( STATUS_SUCCESS);
}

//----------------------------------------------------------------------------------------------
//	X360C_WorkerThreadOut
//----------------------------------------------------------------------------------------------
void X360C_WorkerThreadOut( IN PVOID Context )
{
	//	ϐ錾
	NTSTATUS			Status			= STATUS_SUCCESS;
	PDEVICE_EXTENSION	DeviceExtension	= Context;
	LARGE_INTEGER		TimeOut;
	BOOLEAN				FirstReport		= TRUE;
	OUTPUT_REPORT		LastOutputReport;

	//	Xbh̗D揇ʂ
	KeSetPriorityThread( KeGetCurrentThread(), LOW_REALTIME_PRIORITY );

	while(TRUE)
	{
		//	Xbh̏I`FbN
		if( DeviceExtension->ThreadState == Disable )
		{
			break;
		}

		//	WXgݒǂݍ
		if( DeviceExtension->ResetDevice == TRUE )
		{
			X360C_LoadSetting( DeviceExtension );
			DeviceExtension->ResetRingLight		= TRUE;
			DeviceExtension->ResetDevice		= FALSE;
			FirstReport							= TRUE;
			DeviceExtension->OutputReport.Left	= 0;
			DeviceExtension->OutputReport.Right	= 0;
		}

		//	O Cgݒ肷
		if( DeviceExtension->ResetRingLight == TRUE )
		{
			Status	= X360C_UsbSetRingLight(
						 DeviceExtension
						,(UCHAR)DeviceExtension->Setting.RingLightPattern );
			if( !NT_SUCCESS( Status ) )
			{
				//	16 ms ҋ@
				TimeOut.QuadPart	= -16 * 1000 * 10;
				KeDelayExecutionThread( KernelMode, FALSE, &TimeOut );
			} else {
				DeviceExtension->ResetRingLight	= FALSE;
			}
		}

		//	|[g̓o͂s
		if(	DeviceExtension->Setting.ReduceOutputReport	== 0
		||	FirstReport
		||	DeviceExtension->OutputReport.Left			!= LastOutputReport.Left
		||	DeviceExtension->OutputReport.Right			!= LastOutputReport.Right )
		{
			Status	= X360C_UsbSetReport( DeviceExtension );
			if( !NT_SUCCESS( Status ) )
			{
				//	pCvZbg
				if(	Status	!= STATUS_CANCELLED
				&&	Status	!= STATUS_DEVICE_NOT_CONNECTED )
				{
					X360C_UsbResetPipe( DeviceExtension, DeviceExtension->OutputPipeHandle );
				}
				//	16 ms ҋ@
				TimeOut.QuadPart	= -16 * 1000 * 10;
				KeDelayExecutionThread( KernelMode, FALSE, &TimeOut );
			} else {
				//	ҋ@
				TimeOut.QuadPart	= INTERVAL;
				KeDelayExecutionThread( KernelMode, FALSE, &TimeOut );
			}
			//	Űޔ
			LastOutputReport.Left	= DeviceExtension->OutputReport.Left;
			LastOutputReport.Right	= DeviceExtension->OutputReport.Right;
			FirstReport				= FALSE;
		} else {
			//	ŰɕωȂ 4 ms ҋ@
			TimeOut.QuadPart	= -4 * 1000 * 10;
			KeDelayExecutionThread( KernelMode, FALSE, &TimeOut );
		}
	}

	//	XbhJ
	ObDereferenceObject( DeviceExtension->ThreadObjectOut );
	DeviceExtension->ThreadObjectOut	= NULL;

	//	XbhIƂʒm
	KeSetEvent( &DeviceExtension->ThreadExitingOut, 0 ,FALSE );

	//	XbhI
	PsTerminateSystemThread( STATUS_SUCCESS);
}

//----------------------------------------------------------------------------------------------
//	X360C_DisableDevice
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_DisableDevice( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;
	LONG		Signaled;
	KIRQL		Irql;
	BOOLEAN		Canceled;

	//	[J XbhÎ҂
	if( DeviceExtension->ThreadState == Enable )
	{
		DeviceExtension->ThreadState	= Disable;
		//	foCXAEFCgԂ̏ꍇ Irp LZAʃhCoJ
		KeRaiseIrql( APC_LEVEL, &Irql );
		Signaled	= KeResetEvent( &DeviceExtension->InEvent );
		if( Signaled == 0 )
		{
			IoCancelIrp( DeviceExtension->InIrp );
		}
		KeLowerIrql( Irql );
		//	LZ܂őҋ@
		if( Signaled == 0 )
		{
			KeWaitForSingleObject(
					 &DeviceExtension->InEvent
					,Executive
					,KernelMode
					,FALSE
					,NULL );
		}
		//	͗p[J Xbĥ҂
		KeWaitForSingleObject(
			 &DeviceExtension->ThreadExitingIn
			,Executive
			,KernelMode
			,FALSE
			,NULL );
		if( DeviceExtension->OutputPipeHandle != NULL )
		{
			//	o͗p Irp ĂȂ΃LZ
			KeRaiseIrql( APC_LEVEL, &Irql );
			Signaled	= KeResetEvent( &DeviceExtension->OutEvent );
			if( Signaled == 0 )
			{
				IoCancelIrp( DeviceExtension->OutIrp );
			}
			KeLowerIrql( Irql );
			//	LZ܂őҋ@
			if( Signaled == 0 )
			{
				KeWaitForSingleObject(
						 &DeviceExtension->OutEvent
						,Executive
						,KernelMode
						,FALSE
						,NULL );
			}
			//	o͗p[J Xbĥ҂
			KeWaitForSingleObject(
				 &DeviceExtension->ThreadExitingOut
				,Executive
				,KernelMode
				,FALSE
				,NULL );
		}
	}

	//	hCȍԂ𖳌ɂ
	DeviceExtension->DriverState	= Disable;

	//	QƏICxg
	KeResetEvent( &DeviceExtension->ReferenceEnd );

	//	QƃJE^Z
	DecrementReference( DeviceExtension );
	DecrementReference( DeviceExtension );

	//	SĂ I/O vIAQƏIʒm܂őҋ@
	KeWaitForSingleObject( &DeviceExtension->ReferenceEnd, Executive, KernelMode, FALSE, NULL);

	//	foCX𖳌ɂ
	if( DeviceExtension->DeviceState == Enable )
	{
		//	RtBM[V̑I
		X360C_UsbDeselectConfiguration( DeviceExtension );

		//	foCXLqqpobt@mۂĂΊJ
		if( DeviceExtension->DeviceDescriptor != NULL )
		{
			ExFreePool( DeviceExtension->DeviceDescriptor );
			DeviceExtension->DeviceDescriptor			= NULL;
		}

		//	RtBM[VLqqpobt@mۂĂΊJ
		if( DeviceExtension->ConfigurationDescriptor != NULL )
		{
			ExFreePool( DeviceExtension->ConfigurationDescriptor );
			DeviceExtension->ConfigurationDescriptor	= NULL;
		}

		//	foCX̏Ԃ𖳌ɂ
		DeviceExtension->DeviceState	= Disable;
	}

	//	hCȍԂLɂ
	DeviceExtension->DriverState	= Enable;

	//	QƃJE^̐킹
	IncrementReference( DeviceExtension );
	IncrementReference( DeviceExtension );

	return( Status );
}
