
//	HID Minidriver for Microsoft Xbox 360 Controller

#include	"X360C.h"

//----------------------------------------------------------------------------------------------
//	X360C_UsbSetupPort
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_UsbSetupPort( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;

	//	USB |[g̏Ԃ擾
	Status	= X360C_UsbGetPortStatus( DeviceExtension );
	if( !NT_SUCCESS( Status ) )
	{
		//	USB |[g
		Status	= X360C_UsbResetPort( DeviceExtension );
		if( !NT_SUCCESS( Status ) )
		{
			return( Status );
		}

		//	ēx USB |[g̏Ԃ擾
		Status	= X360C_UsbGetPortStatus( DeviceExtension );
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_UsbGetPortStatus
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_UsbGetPortStatus( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;
	ULONG		PortStatus;

	//	USB |[g̏Ԃ擾
	Status	= X360C_CallUSBDI(
				 DeviceExtension
				,IOCTL_INTERNAL_USB_GET_PORT_STATUS
				,&PortStatus );
	if( !NT_SUCCESS( Status ) )
	{
		return( Status );
	}

	//	foCX USB |[gɐڑĂ邩ׂ
	if( ( PortStatus & USBD_PORT_CONNECTED ) != USBD_PORT_CONNECTED )
	{
		return( STATUS_NO_SUCH_DEVICE );
	}

	//	USB |[gLׂ
	if( ( PortStatus & USBD_PORT_ENABLED ) != USBD_PORT_ENABLED )
	{
		return( STATUS_DEVICE_NOT_READY );
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_UsbResetPort
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_UsbResetPort( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	USB |[gZbg
	return( X360C_CallUSBDI( DeviceExtension, IOCTL_INTERNAL_USB_RESET_PORT, NULL ) );
}

//----------------------------------------------------------------------------------------------
//	X360C_UsbResetPipe
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_UsbResetPipe(
	 IN PDEVICE_EXTENSION	DeviceExtension
	,IN USBD_PIPE_HANDLE	PipeHandle )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;
	USHORT		UrbSize	= sizeof( struct _URB_PIPE_REQUEST );
	PURB		Urb;

	//	URB pobt@쐬
	Urb	= (PURB)ExAllocatePool( NonPagedPool, UrbSize );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	pCvZbg邽߂ URB 쐬
	Urb->UrbHeader.Length			= (USHORT) sizeof(struct _URB_PIPE_REQUEST);
	Urb->UrbHeader.Function			= URB_FUNCTION_RESET_PIPE;
	Urb->UrbPipeRequest.PipeHandle	= PipeHandle;

	//	pCvZbg邽߂ URB 𔭍s
	Status	= X360C_CallUSBDI( DeviceExtension, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
	if( !NT_SUCCESS( Status ) || !USBD_SUCCESS( Urb->UrbHeader.Status ) )
	{
		Status	= STATUS_UNSUCCESSFUL;
	}

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_UsbGetDeviceDescriptor
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_UsbGetDeviceDescriptor( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS				Status	= STATUS_SUCCESS;
	USHORT					UrbSize	= sizeof( struct _URB_CONTROL_DESCRIPTOR_REQUEST );
	PURB					Urb;
	LONG					DeviceDescriptorSize
									= sizeof( USB_DEVICE_DESCRIPTOR );
	PUSB_DEVICE_DESCRIPTOR	DeviceDescriptor;

	//	URB pobt@쐬
	Urb	= (PURB)ExAllocatePool( NonPagedPool, UrbSize );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	foCXLqqpobt@쐬
	DeviceDescriptor	= ExAllocatePool( NonPagedPool, DeviceDescriptorSize );
	if( DeviceDescriptor == NULL )
	{
		ExFreePool( Urb );
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	foCXLqq擾ׂ URB 쐬
	UsbBuildGetDescriptorRequest(
		 Urb
		,UrbSize
		,USB_DEVICE_DESCRIPTOR_TYPE
		,0
		,0
		,DeviceDescriptor
		,NULL
		,DeviceDescriptorSize
		,NULL );

	//	foCXLqq擾ׂ URB 𔭍s
	Status	= X360C_CallUSBDI( DeviceExtension, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
	if( !NT_SUCCESS( Status ) || !USBD_SUCCESS( Urb->UrbHeader.Status ) )
	{
		Status	= STATUS_UNSUCCESSFUL;
	}

	//	foCXLqq擾
	DeviceExtension->DeviceDescriptor	= DeviceDescriptor;

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_UsbGetConfigurationDescriptor
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_UsbGetConfigurationDescriptor( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS						Status	= STATUS_SUCCESS;
	USHORT							UrbSize	= sizeof( struct _URB_CONTROL_DESCRIPTOR_REQUEST );
	PURB							Urb;
	LONG							ConfigurationDescriptorSize
											= sizeof( USB_CONFIGURATION_DESCRIPTOR );
	PUSB_CONFIGURATION_DESCRIPTOR	ConfigurationDescriptor;

	//	URB pobt@쐬
	Urb	= (PURB)ExAllocatePool( NonPagedPool, UrbSize );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	RtBM[VLqq̓TCYmȂ̂ŒKv
	while( TRUE )
	{
		//	RtBM[VLqqpobt@쐬
		ConfigurationDescriptor	= ExAllocatePool( NonPagedPool, ConfigurationDescriptorSize );
		if( ConfigurationDescriptor == NULL )
		{
			ExFreePool( Urb );
			return( STATUS_INSUFFICIENT_RESOURCES );
		}

		//	RtBM[VLqq擾ׂ URB 쐬
		UsbBuildGetDescriptorRequest(
			 Urb
			,UrbSize
			,USB_CONFIGURATION_DESCRIPTOR_TYPE
			,0
			,0
			,ConfigurationDescriptor
			,NULL
			,ConfigurationDescriptorSize
			,NULL );

		//	RtBM[VLqq擾ׂ URB 𔭍s
		Status	= X360C_CallUSBDI( DeviceExtension, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
		if( !NT_SUCCESS( Status ) || !USBD_SUCCESS( Urb->UrbHeader.Status ) )
		{
			Status	= STATUS_UNSUCCESSFUL;
			break;
		}

		//	쐬RtBM[VLqq̃TCYꍇAč쐬
		if( ConfigurationDescriptor->wTotalLength > ConfigurationDescriptorSize )
		{
			ConfigurationDescriptorSize	= ConfigurationDescriptor->wTotalLength;
			ExFreePool( ConfigurationDescriptor );
			ConfigurationDescriptor		= NULL;
		} else {
			break;
		}
	}

	//	RtBM[VLqq擾
	DeviceExtension->ConfigurationDescriptor	= ConfigurationDescriptor;

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_UsbGetInterfaceDescriptor
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_UsbGetInterfaceDescriptor( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS					Status	= STATUS_SUCCESS;
	LONG						Index;
	PUSB_INTERFACE_DESCRIPTOR	InterfaceDescriptor	=
		(PUSB_INTERFACE_DESCRIPTOR)DeviceExtension->ConfigurationDescriptor;

	for( Index = 0; Index < DeviceExtension->ConfigurationDescriptor->bNumInterfaces; Index ++ )
	{
		//	C^[tFCXLqq擾
		InterfaceDescriptor	= USBD_ParseConfigurationDescriptorEx(
								 DeviceExtension->ConfigurationDescriptor
								,InterfaceDescriptor
								,Index
								,0
								,-1
								,-1
								,-1 );
		if( InterfaceDescriptor == NULL )
		{
			return( STATUS_INSUFFICIENT_RESOURCES );
		}

		//	C^[tFCXLqq擾
		DeviceExtension->InterfaceDescriptor[Index]	= InterfaceDescriptor;
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_UsbSelectConfiguration
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_UsbSelectConfiguration( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS					Status	= STATUS_SUCCESS;
	USBD_INTERFACE_LIST_ENTRY	InterfaceListEntry[5];
	PURB						Urb;
	PUSBD_INTERFACE_INFORMATION	InterfaceInformation;
	PUSBD_PIPE_INFORMATION		PipeInformation;

	//	C^[tFCXݒ肷
	InterfaceListEntry[0].InterfaceDescriptor	= DeviceExtension->InterfaceDescriptor[0];
	InterfaceListEntry[0].Interface				= NULL;
	InterfaceListEntry[1].InterfaceDescriptor	= DeviceExtension->InterfaceDescriptor[1];
	InterfaceListEntry[1].Interface				= NULL;
	InterfaceListEntry[2].InterfaceDescriptor	= DeviceExtension->InterfaceDescriptor[2];
	InterfaceListEntry[2].Interface				= NULL;
	InterfaceListEntry[3].InterfaceDescriptor	= DeviceExtension->InterfaceDescriptor[3];
	InterfaceListEntry[3].Interface				= NULL;
	InterfaceListEntry[4].InterfaceDescriptor	= NULL;
	InterfaceListEntry[4].Interface				= NULL;

	//	RtBM[VI邽߂ URB 쐬
	Urb	= USBD_CreateConfigurationRequestEx(
			 DeviceExtension->ConfigurationDescriptor
			,InterfaceListEntry );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	RtBM[VIׂ URB 𔭍s
	Status	= X360C_CallUSBDI( DeviceExtension, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
	if( NT_SUCCESS( Status ) )
	{
		//	C^[tFCX CtH[V擾
		InterfaceInformation					= &(Urb->UrbSelectConfiguration.Interface);
		DeviceExtension->InterfaceInformation	= InterfaceInformation;

		//	RtBM[V nh擾
		DeviceExtension->ConfigurationHandle
							= Urb->UrbSelectConfiguration.ConfigurationHandle;

		//	pCv擾
		switch( InterfaceInformation->NumberOfPipes )
		{
			case 0:
				ExFreePool( Urb );
				return( STATUS_INSUFFICIENT_RESOURCES );
				break;

			case 1:
				PipeInformation	= &InterfaceInformation->Pipes[0];
				DeviceExtension->InputPipeHandle	= PipeInformation->PipeHandle;
				DeviceExtension->OutputPipeHandle	= NULL;
				break;

			default:
				PipeInformation	= &InterfaceInformation->Pipes[0];
				DeviceExtension->InputPipeHandle	= PipeInformation->PipeHandle;
				PipeInformation	= &InterfaceInformation->Pipes[1];
				DeviceExtension->OutputPipeHandle	= PipeInformation->PipeHandle;
				break;
		}

	} else {
		ExFreePool( Urb );
		return( STATUS_UNSUCCESSFUL );
	}

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_UsbDeselectConfiguration
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_UsbDeselectConfiguration( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS	Status	= STATUS_SUCCESS;
	USHORT		UrbSize	= sizeof( struct _URB_CONTROL_DESCRIPTOR_REQUEST );
	PURB		Urb;

	//	URB pobt@쐬
	Urb	= (PURB)ExAllocatePool( NonPagedPool, UrbSize );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	RtBM[VIׂ URB 쐬
	UsbBuildSelectConfigurationRequest( Urb, UrbSize, NULL );

	//	RtBM[VIׂ URB 𔭍s
	Status	= X360C_CallUSBDI( DeviceExtension, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
	if( !NT_SUCCESS( Status ) || !USBD_SUCCESS( Urb->UrbHeader.Status ) )
	{
		ExFreePool( Urb );
		return( STATUS_UNSUCCESSFUL );
	}

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_UsbGetReport
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_UsbGetReport( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS	Status		= STATUS_SUCCESS;
	USHORT		UrbSize		= sizeof( struct _URB_BULK_OR_INTERRUPT_TRANSFER );
	PURB		Urb;
	LONG		BufferSize	= sizeof( INPUT_REPORT );
	UCHAR *		Buffer;
	KIRQL		Irql;

	//	URB pobt@쐬
	Urb	= (PURB)ExAllocatePool( NonPagedPool, UrbSize );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	̓|[gpobt@쐬
	Buffer	= (UCHAR *)ExAllocatePool( NonPagedPool, BufferSize );
	if( Buffer == NULL )
	{
		ExFreePool( Urb );
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	̓|[g擾ׂ URB 쐬
	UsbBuildInterruptOrBulkTransferRequest(
		 Urb
		,UrbSize
		,DeviceExtension->InputPipeHandle
		,Buffer
		,NULL
		,BufferSize
		,USBD_TRANSFER_DIRECTION_IN | USBD_SHORT_TRANSFER_OK
		,NULL );

	//	̓|[g擾ׂ URB 𔭍s
	Status	= X360C_CallUSBDIIn( DeviceExtension, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
	if( !NT_SUCCESS( Status ) || !USBD_SUCCESS( Urb->UrbHeader.Status ) )
	{
		ExFreePool( Buffer );
		ExFreePool( Urb );
		return( STATUS_UNSUCCESSFUL );
	}

	//	̓|[g擾
	if( Status != STATUS_TIMEOUT )
	{
		//	|[gmF
		if( Buffer[1] == 20 )
		{
			KeAcquireSpinLock( &DeviceExtension->LockReport, &Irql );
			RtlCopyMemory( &DeviceExtension->InputReport, Buffer, BufferSize );
			KeReleaseSpinLock( &DeviceExtension->LockReport, Irql );
		}
	}

	//	̓|[gpobt@
	ExFreePool( Buffer );

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_UsbSetReport
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_UsbSetReport( IN PDEVICE_EXTENSION DeviceExtension )
{
	//	ϐ錾
	NTSTATUS	Status		= STATUS_SUCCESS;
	USHORT		UrbSize		= sizeof( struct _URB_BULK_OR_INTERRUPT_TRANSFER );
	PURB		Urb;
	LONG		BufferSize	= sizeof( OUTPUT_REPORT );
	UCHAR *		Buffer;
	KIRQL		Irql;

	//	URB pobt@쐬
	Urb	= (PURB)ExAllocatePool( NonPagedPool, UrbSize );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	o̓|[gpobt@쐬
	Buffer	= (UCHAR *)ExAllocatePool( NonPagedPool, BufferSize );
	if( Buffer == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	o̓|[gݒ肷
	KeAcquireSpinLock( &DeviceExtension->LockReport, &Irql);
	RtlCopyMemory( Buffer, &DeviceExtension->OutputReport, BufferSize );
	KeReleaseSpinLock( &DeviceExtension->LockReport, Irql);

	//	o̓|[gݒ肷ׂ URB 쐬
	UsbBuildInterruptOrBulkTransferRequest(
		 Urb
		,UrbSize
		,DeviceExtension->OutputPipeHandle
		,Buffer
		,NULL
		,BufferSize
		,USBD_SHORT_TRANSFER_OK
		,NULL );
	//	o̓|[gݒ肷ׂ URB 𔭍s
	Status	= X360C_CallUSBDIOut( DeviceExtension, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
	if( !NT_SUCCESS( Status ) || !USBD_SUCCESS( Urb->UrbHeader.Status ) )
	{
		Status	= STATUS_UNSUCCESSFUL;
	}

	//	o̓|[gpobt@
	ExFreePool( Buffer );

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_UsbSetRingLight
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_UsbSetRingLight(
	 IN PDEVICE_EXTENSION	DeviceExtension
	,IN UCHAR				State )
{
	//	ϐ錾
	NTSTATUS	Status		= STATUS_SUCCESS;
	USHORT		UrbSize		= sizeof( struct _URB_BULK_OR_INTERRUPT_TRANSFER );
	PURB		Urb;
	LONG		BufferSize	= 3;
	UCHAR *		Buffer;
	KIRQL		Irql;

	//	URB pobt@쐬
	Urb	= (PURB)ExAllocatePool( NonPagedPool, UrbSize );
	if( Urb == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	o̓|[gpobt@쐬
	Buffer	= (UCHAR *)ExAllocatePool( NonPagedPool, BufferSize );
	if( Buffer == NULL )
	{
		return( STATUS_INSUFFICIENT_RESOURCES );
	}

	//	o̓|[gݒ肷
	Buffer[0]	= 1;
	Buffer[1]	= 3;
	Buffer[2]	= State;

	//	o̓|[gݒ肷ׂ URB 쐬
	UsbBuildInterruptOrBulkTransferRequest(
		 Urb
		,UrbSize
		,DeviceExtension->OutputPipeHandle
		,Buffer
		,NULL
		,BufferSize
		,USBD_SHORT_TRANSFER_OK
		,NULL );
	//	o̓|[gݒ肷ׂ URB 𔭍s
	Status	= X360C_CallUSBDIOut( DeviceExtension, IOCTL_INTERNAL_USB_SUBMIT_URB, Urb );
	if( !NT_SUCCESS( Status ) || !USBD_SUCCESS( Urb->UrbHeader.Status ) )
	{
		Status	= STATUS_UNSUCCESSFUL;
	}

	//	o̓|[gpobt@
	ExFreePool( Buffer );

	//	쐬 URB 
	ExFreePool( Urb );

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_CallUSBDI
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_CallUSBDI(
	 IN PDEVICE_EXTENSION	DeviceExtension
	,IN ULONG				IOCTLCode
	,IN OUT PVOID			Arg1 )
{
	//	ϐ錾
	NTSTATUS			Status	= STATUS_SUCCESS;
	PIRP				Irp;
	KEVENT				WaitEvent;
	IO_STATUS_BLOCK		IoStatus;
	PIO_STACK_LOCATION	NextIrpStack;

	//	I/O v쐬
	KeInitializeEvent( &WaitEvent, NotificationEvent, FALSE );
	Irp	= IoBuildDeviceIoControlRequest(
			 IOCTLCode
			,DeviceExtension->NextDeviceObject
			,NULL
			,0
			,NULL
			,0
			,TRUE
			,&WaitEvent
			,&IoStatus );
	NextIrpStack	= IoGetNextIrpStackLocation( Irp );
	NextIrpStack->Parameters.Others.Argument1	= Arg1;

	//	ʃhCo I/O v𔭍s
	Status	= IoCallDriver( DeviceExtension->NextDeviceObject, Irp );
	if( Status == STATUS_PENDING )
	{
		//	ʃhCȍ܂őҋ@
		KeWaitForSingleObject(
			 &WaitEvent
			,Executive
			,KernelMode
			,FALSE
			,NULL );
		Status	= IoStatus.Status;
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_CallUSBDIIn
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_CallUSBDIIn(
	 IN PDEVICE_EXTENSION	DeviceExtension
	,IN ULONG				IOCTLCode
	,IN OUT PVOID			Arg1 )
{
	//	ϐ錾
	NTSTATUS			Status	= STATUS_SUCCESS;
	IO_STATUS_BLOCK		IoStatus;
	PIO_STACK_LOCATION	NextIrpStack;

	//	I/O v쐬
	KeResetEvent( &DeviceExtension->InEvent );
	DeviceExtension->InIrp	= IoBuildDeviceIoControlRequest(
			 IOCTLCode
			,DeviceExtension->NextDeviceObject
			,NULL
			,0
			,NULL
			,0
			,TRUE
			,&DeviceExtension->InEvent
			,&IoStatus );
	NextIrpStack	= IoGetNextIrpStackLocation( DeviceExtension->InIrp );
	NextIrpStack->Parameters.Others.Argument1	= Arg1;

	//	ʃhCo I/O v𔭍s
	Status	= IoCallDriver( DeviceExtension->NextDeviceObject, DeviceExtension->InIrp );
	if( Status == STATUS_PENDING )
	{
		//	ʃhCȍ܂őҋ@
		KeWaitForSingleObject(
			 &DeviceExtension->InEvent
			,Executive
			,KernelMode
			,FALSE
			,NULL );
		Status	= IoStatus.Status;
	}

	return( Status );
}

//----------------------------------------------------------------------------------------------
//	X360C_CallUSBDIOut
//----------------------------------------------------------------------------------------------
NTSTATUS X360C_CallUSBDIOut(
	 IN PDEVICE_EXTENSION	DeviceExtension
	,IN ULONG				IOCTLCode
	,IN OUT PVOID			Arg1 )
{
	//	ϐ錾
	NTSTATUS			Status	= STATUS_SUCCESS;
	IO_STATUS_BLOCK		IoStatus;
	PIO_STACK_LOCATION	NextIrpStack;

	//	I/O v쐬
	KeResetEvent( &DeviceExtension->OutEvent );
	DeviceExtension->OutIrp	= IoBuildDeviceIoControlRequest(
			 IOCTLCode
			,DeviceExtension->NextDeviceObject
			,NULL
			,0
			,NULL
			,0
			,TRUE
			,&DeviceExtension->OutEvent
			,&IoStatus );
	NextIrpStack	= IoGetNextIrpStackLocation( DeviceExtension->OutIrp );
	NextIrpStack->Parameters.Others.Argument1	= Arg1;

	//	ʃhCo I/O v𔭍s
	Status	= IoCallDriver( DeviceExtension->NextDeviceObject, DeviceExtension->OutIrp );
	if( Status == STATUS_PENDING )
	{
		//	ʃhCȍ܂őҋ@
		KeWaitForSingleObject(
			 &DeviceExtension->OutEvent
			,Executive
			,KernelMode
			,FALSE
			,NULL );
		Status	= IoStatus.Status;
	}

	return( Status );
}
