
//	HID Minidriver for Microsoft Xbox 360 Controller

#include	<wdm.h>
#include	<hidport.h>
#include	<usbdi.h>
#include	<usbdlib.h>

//----------------------------------------------------------------------------------------------
//	萔錾
//----------------------------------------------------------------------------------------------

#define	Enable					TRUE
#define	Disable					FALSE

#define	INPUT_REPORT_SIZE		20
#define	OUTPUT_REPORT_SIZE		3

#define	INTERVAL				-1

#define	DIRECTIONAL_PAD_UP		0x0001
#define	DIRECTIONAL_PAD_DOWN	0x0002
#define	DIRECTIONAL_PAD_LEFT	0x0004
#define	DIRECTIONAL_PAD_RIGHT	0x0008
#define	START_BUTTON			0x0010
#define	BACK_BUTTON				0x0020
#define	LEFT_STICK_BUTTON		0x0040
#define	RIGHT_STICK_BUTTON		0x0080
#define	LB_BUTTON				0x0100
#define	RB_BUTTON				0x0200
#define	XBOX_GUIDE_BUTTON		0x0400
#define	A_BUTTON				0x1000
#define	B_BUTTON				0x2000
#define	X_BUTTON				0x4000
#define	Y_BUTTON				0x8000

#define	REGISTRY_PATH			L"\\Registry\\Machine\\System\\CurrentControlSet\\Services\\Xbox360Controller\\Parameters"
#define	CURRENT_SETTING			L"Current Setting"

//----------------------------------------------------------------------------------------------
//	\̐錾
//----------------------------------------------------------------------------------------------

//	WXgݒ
typedef struct _SETTING
{
	ULONG	StickUsage[2];
	ULONG	StickReverse[2];
	ULONG	StickMinThreshold;
	ULONG	StickMaxThreshold;
	ULONG	StickHatSwitchThreshold;
	ULONG	DirectionalPadUsage;
	ULONG	DirectionalPadReverse;
	ULONG	DirectionalPadWayUsage[4];
	ULONG	DirectionalPadWayRapidFire[4];
	ULONG	TriggerUsage[2];
	ULONG	TriggerRapidFire[2];
	ULONG	TriggerMinThreshold;
	ULONG	TriggerMaxThreshold;
	ULONG	TriggerButtonThreshold;
	ULONG	ButtonUsage[11];
	ULONG	ButtonRapidFire[11];
	ULONG	ButtonRapidFireSpeed[11];
	ULONG	RingLightPattern;
	ULONG	ForceFeedback;
	ULONG	ActuatorLevel[2];
	ULONG	ActuatorMinThreshold;
	ULONG	ActuatorMaxThreshold;
	ULONG	StickDeadZoneType;
	ULONG	SlideSliderValue[4];
	ULONG	SlideCenterValue[4];
	ULONG	ReverseSliderValue[4];
	ULONG	LimitationMode;
	ULONG	DevicesArePolled;
	ULONG	ReduceOutputReport;
} SETTING, *PSETTING;

//	̓|[g
typedef struct _INPUT_REPORT
{
	UCHAR	Zero1;
	UCHAR	Length;
	USHORT	DigitalButton;
	UCHAR	LeftTrigger;
	UCHAR	RightTrigger;
	SHORT	LeftStickX;
	SHORT	LeftStickY;
	SHORT	RightStickX;
	SHORT	RightStickY;
	UCHAR	Dummy1;
	UCHAR	Dummy2;
	UCHAR	Dummy3;
	UCHAR	Dummy4;
	UCHAR	Dummy5;
	UCHAR	Dummy6;
} INPUT_REPORT, *PINPUT_REPORT;

//	o̓|[g
typedef struct _OUTPUT_REPORT
{
	UCHAR	Zero1;
	UCHAR	Length;
	UCHAR	Zero2;
	UCHAR	Left;
	UCHAR	Right;
	UCHAR	Zero3;
	UCHAR	Zero4;
	UCHAR	Zero5;
} OUTPUT_REPORT, *POUTPUT_REPORT;

//	|[g
typedef struct _REPORT
{
#if( WINVER < 0x0500 )
	USHORT	XAxis;
	USHORT	YAxis;
	USHORT	XRotation;
	USHORT	YRotation;
#else
	SHORT	XAxis;
	SHORT	YAxis;
	SHORT	XRotation;
	SHORT	YRotation;
#endif
	UCHAR	ZAxis;
	UCHAR	ZRotation;
	UCHAR	Slider2;
	UCHAR	Slider1;
	CHAR	HatSwitch1;
	CHAR	HatSwitch2;
	CHAR	HatSwitch3;
	CHAR	HatSwitch4;
	ULONG	Button;
} REPORT, *PREPORT;

//	foCX GNXeV
typedef struct _DEVICE_EXTENSION
{
	//	I/O v
	ULONG							DriverState;
	ULONG							DeviceState;
	LONG							ReferenceCount;
	KEVENT							ReferenceEnd;

	//	[J Xbh
	BOOLEAN							ThreadState;
	PVOID							ThreadObjectIn;
	KEVENT							ThreadExitingIn;
	PVOID							ThreadObjectOut;
	KEVENT							ThreadExitingOut;

	//	USB
	PDEVICE_OBJECT					NextDeviceObject;
	PUSB_DEVICE_DESCRIPTOR			DeviceDescriptor;
	PUSB_CONFIGURATION_DESCRIPTOR	ConfigurationDescriptor;
	PUSB_INTERFACE_DESCRIPTOR		InterfaceDescriptor[4];
	PUSBD_INTERFACE_INFORMATION		InterfaceInformation;
	USBD_CONFIGURATION_HANDLE		ConfigurationHandle;
	USBD_PIPE_HANDLE				InputPipeHandle;
	USBD_PIPE_HANDLE				OutputPipeHandle;
	KEVENT							InEvent;
	PIRP							InIrp;
	KEVENT							OutEvent;
	PIRP							OutIrp;

	//	|[g
	INPUT_REPORT					InputReport;
	OUTPUT_REPORT					OutputReport;
	KSPIN_LOCK						LockReport;

	//	WXgݒ
	BOOLEAN							ResetDevice;
	SETTING							Setting;

	//	O Cgݒ
	BOOLEAN							ResetRingLight;

	//	A
	ULONG							TriggerLastState[2];
	LONG							TriggerRapidFireCount[2];
	ULONG							DirectionalPadWayLastState[4];
	LONG							DirectionalPadWayRapidFireCount[4];
	ULONG							ButtonLastState[11];
	LONG							ButtonRapidFireCount[11];
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;

//	cobp
typedef struct _READ_TIMER
{
	KDPC				ReadTimerDpc;
	KTIMER				ReadTimer;
	PIRP				Irp;
	PDEVICE_EXTENSION	DeviceExtension;
} READ_TIMER, *PREAD_TIMER;

//----------------------------------------------------------------------------------------------
//	}N`
//----------------------------------------------------------------------------------------------

#define	GET_PHYSICAL_DEVICE_OBJECT( DO ) \
		( ((PHID_DEVICE_EXTENSION)(DO)->DeviceExtension)->PhysicalDeviceObject )

#define	GET_NEXT_DEVICE_OBJECT( DO ) \
		( ((PHID_DEVICE_EXTENSION)(DO)->DeviceExtension)->NextDeviceObject )

#define GET_MINIDRIVER_DEVICE_EXTENSION( DO ) \
		( (PDEVICE_EXTENSION) \
		(((PHID_DEVICE_EXTENSION)(DO)->DeviceExtension)->MiniDeviceExtension) )

#ifndef min
	#define min( D0, D1 )( D0 < D1 ? D0 : D1 )
#endif
#ifndef max
	#define max( D0, D1 )( D0 < D1 ? D1 : D0 )
#endif

//----------------------------------------------------------------------------------------------
//	O[oϐ
//----------------------------------------------------------------------------------------------

extern BOOLEAN	DevicesArePolled;
extern LONG		LimitationMode;

//----------------------------------------------------------------------------------------------
//	֐錾
//----------------------------------------------------------------------------------------------

//	AddDevice.c
NTSTATUS X360C_AddDevice( IN PDRIVER_OBJECT, IN PDEVICE_OBJECT );

//	Unload.c
VOID X360C_Unload( IN PDRIVER_OBJECT );

//	InternalDeviceControl.c
NTSTATUS X360C_InternalDeviceControl( IN PDEVICE_OBJECT, IN PIRP );
NTSTATUS X360C_IOCTLHIDGetDeviceAttributes( IN PDEVICE_EXTENSION, IN PIRP, IN PIO_STACK_LOCATION );
NTSTATUS X360C_IOCTLHIDGetDeviceDescriptor( IN PDEVICE_EXTENSION, IN PIRP, IN PIO_STACK_LOCATION );
NTSTATUS X360C_IOCTLHIDGetReportDescriptor( IN PDEVICE_EXTENSION, IN PIRP, IN PIO_STACK_LOCATION );
NTSTATUS X360C_IOCTLHIDReadReport( IN PDEVICE_EXTENSION, IN PIRP, IN PIO_STACK_LOCATION );
VOID X360C_ReadTimerDpcRoutine( IN PKDPC, IN PVOID, IN PVOID, IN PVOID );
NTSTATUS X360C_IOCTLHIDWriteReport( IN PDEVICE_EXTENSION, IN PIRP, IN PIO_STACK_LOCATION );
NTSTATUS X360C_IOCTLHIDGetFeature( IN PDEVICE_EXTENSION, IN PIRP, IN PIO_STACK_LOCATION );

//	GenerateReport.c
NTSTATUS X360C_GenerateInputReport( IN PDEVICE_EXTENSION, OUT REPORT * );
VOID X360C_AdjustStick_Axis( LONG, SHORT *, SHORT *, LONG, LONG );
VOID X360C_AdjustStick_HatSwitch( UCHAR *, SHORT, SHORT, LONG );
VOID X360C_AdjustSlider_Slider( UCHAR *, LONG, LONG );
VOID X360C_AdjustSlider_Button( ULONG *, UCHAR, LONG );
VOID X360C_SetAxisRotation( SHORT *, SHORT, SHORT, ULONG, ULONG );
VOID X360C_SetHatSwitch( UCHAR *, UCHAR, ULONG, ULONG );
VOID X360C_SetSlider( SHORT *, SHORT, ULONG, ULONG, ULONG, ULONG *, ULONG * );
VOID X360C_SetButton( ULONG *, ULONG, ULONG, ULONG, ULONG, ULONG *, ULONG * );
NTSTATUS X360C_GenerateOutputReport( IN PDEVICE_EXTENSION, OUT UCHAR * );
VOID X360C_SetActuator( UCHAR *, UCHAR, LONG, LONG, LONG );
NTSTATUS X360C_LoadSetting( IN PDEVICE_EXTENSION );

//	Math.c
LONG Sin( LONG );
LONG Cos( LONG );
LONG ATan2( LONG, LONG );
LONG Sqrt( LONG );

//	PnP.c
NTSTATUS X360C_PnP( IN PDEVICE_OBJECT, IN PIRP );
NTSTATUS X360C_StartDevice( IN PDEVICE_OBJECT, IN PDEVICE_EXTENSION, IN PIRP );
VOID X360C_WorkerThreadIn( IN PVOID );
VOID X360C_WorkerThreadOut( IN PVOID );
NTSTATUS X360C_QueryCapabilities( IN PDEVICE_OBJECT, IN PIRP, IN PIO_STACK_LOCATION );
NTSTATUS X360C_SurpriseRemoval( IN PDEVICE_OBJECT, IN PDEVICE_EXTENSION, IN PIRP );
NTSTATUS X360C_RemoveDevice( IN PDEVICE_OBJECT, IN PDEVICE_EXTENSION, IN PIRP );
NTSTATUS X360C_EnableDevice( IN PDEVICE_OBJECT, IN PDEVICE_EXTENSION, IN PIRP );
NTSTATUS X360C_DisableDevice( IN PDEVICE_EXTENSION );

//	Power.c
NTSTATUS X360C_Power( IN PDEVICE_OBJECT, IN PIRP );

//	Common.c
VOID CompleteRequest( IN PIRP, IN NTSTATUS, IN ULONG );
NTSTATUS SendRequest( IN PDEVICE_OBJECT, IN PIRP );
NTSTATUS SendAndWaitRequest( IN PDEVICE_OBJECT, IN PIRP );
NTSTATUS IoCompletionRoutine( IN PDEVICE_OBJECT, IN PIRP, IN PKEVENT );
NTSTATUS IncrementReference( IN PDEVICE_EXTENSION );
NTSTATUS DecrementReference( IN PDEVICE_EXTENSION );
ULONG GetRegistryLong( IN PWSTR, IN PWSTR, IN ULONG );
ULONG GetRegistryBinary( IN PWSTR, IN PWSTR, IN CHAR *, IN LONG );

//	USB.c
NTSTATUS X360C_UsbSetupPort( IN PDEVICE_EXTENSION );
NTSTATUS X360C_UsbGetPortStatus( IN PDEVICE_EXTENSION );
NTSTATUS X360C_UsbResetPort( IN PDEVICE_EXTENSION );
NTSTATUS X360C_UsbResetPipe( IN PDEVICE_EXTENSION, IN USBD_PIPE_HANDLE );
NTSTATUS X360C_UsbGetDeviceDescriptor( IN PDEVICE_EXTENSION );
NTSTATUS X360C_UsbGetConfigurationDescriptor( IN PDEVICE_EXTENSION );
NTSTATUS X360C_UsbGetInterfaceDescriptor( IN PDEVICE_EXTENSION );
NTSTATUS X360C_UsbSelectConfiguration( IN PDEVICE_EXTENSION );
NTSTATUS X360C_UsbDeselectConfiguration( IN PDEVICE_EXTENSION );
NTSTATUS X360C_UsbGetReport( IN PDEVICE_EXTENSION );
NTSTATUS X360C_UsbSetReport( IN PDEVICE_EXTENSION );
NTSTATUS X360C_UsbSetRingLight( IN PDEVICE_EXTENSION, UCHAR );
NTSTATUS X360C_CallUSBDI( IN PDEVICE_EXTENSION, IN ULONG, IN OUT PVOID );
NTSTATUS X360C_CallUSBDIIn( IN PDEVICE_EXTENSION, IN ULONG, IN OUT PVOID );
NTSTATUS X360C_CallUSBDIOut( IN PDEVICE_EXTENSION, IN ULONG, IN OUT PVOID );
