
//	Force Feedback Driver for Microsoft Xbox 360 Controller

#include	"X360CFF.h"

//----------------------------------------------------------------------------------------------
//	O[oϐ錾
//----------------------------------------------------------------------------------------------

IDirectInputEffectDriverVtbl	EffectDriverVtbl	=
{
	 EffectDriver_QueryInterface
	,EffectDriver_AddRef
	,EffectDriver_Release
	,EffectDriver_DeviceID
	,EffectDriver_GetVersions
	,EffectDriver_Escape
	,EffectDriver_SetGain
	,EffectDriver_SendForceFeedbackCommand
	,EffectDriver_GetForceFeedbackState
	,EffectDriver_DownloadEffect
	,EffectDriver_DestroyEffect
	,EffectDriver_StartEffect
	,EffectDriver_StopEffect
	,EffectDriver_GetEffectStatus
};

//----------------------------------------------------------------------------------------------
//	EffectDriver_New
//----------------------------------------------------------------------------------------------

STDMETHODIMP EffectDriver_New(
	 REFIID		InterfaceID
	,LPVOID *	Obj )
{
	//	ϐ錾
	HRESULT			Result	= S_OK;
	EFFECT_DRIVER *	EffectDriver;

	//	GtFNg hCo[쐬
	EffectDriver	= LocalAlloc( LPTR | LMEM_ZEROINIT, sizeof( EFFECT_DRIVER ) );
	if( EffectDriver != NULL )
	{
		//	C^[tFCXݒ肷
		EffectDriver->Interface.lpVtbl	= &EffectDriverVtbl;

		//	QƃJE^
		EffectDriver->ReferenceCount	= 1;
		IncrementReference();

		//	GtFNg hCȍlݒ肷
		EffectDriver->EffectCount	= 1;
		EffectDriver->Actuators		= TRUE;

		//	g̃C^[tFCX擾
		Result	= EffectDriver_QueryInterface( &EffectDriver->Interface, InterfaceID, Obj );
		//	QƃJE^𒲐
		EffectDriver_Release( &EffectDriver->Interface );
	} else {
		Result	= E_OUTOFMEMORY;
	}

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_QueryInterface
//----------------------------------------------------------------------------------------------
STDMETHODIMP EffectDriver_QueryInterface(
	 IDirectInputEffectDriver *	EffectDriverInterface
	,REFIID						InterfaceID
	,LPVOID *					Obj )
{
	//	ϐ錾
	HRESULT	Result	= S_OK;

	//	g̃C^[tFCXԂ
	if(	( IsEqualIID( InterfaceID, &IID_IUnknown ) )
	||	( IsEqualIID( InterfaceID, &IID_IDirectInputEffectDriver ) ) )
	{
		EffectDriver_AddRef( EffectDriverInterface );
		*Obj	= EffectDriverInterface;
	} else {
		*Obj	= NULL;
		Result	= E_NOINTERFACE;
	}

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_AddRef
//----------------------------------------------------------------------------------------------
STDMETHODIMP_(ULONG) EffectDriver_AddRef( IDirectInputEffectDriver * EffectDriverInterface )
{
	//	ϐ錾
	HRESULT			Result			= S_OK;
	EFFECT_DRIVER *	EffectDriver	= (PEFFECT_DRIVER)EffectDriverInterface;

	//	QƃJE^Z
	InterlockedIncrement( (LPLONG)&EffectDriver->ReferenceCount );

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_Release
//----------------------------------------------------------------------------------------------
STDMETHODIMP_(ULONG) EffectDriver_Release( IDirectInputEffectDriver * EffectDriverInterface )
{
	//	ϐ錾
	ULONG			Result			= 0;
	EFFECT_DRIVER *	EffectDriver	= (PEFFECT_DRIVER)EffectDriverInterface;
	DWORD			ExitCode;

	//	QƃJE^Z
	Result	= InterlockedDecrement( (LPLONG)&EffectDriver->ReferenceCount );
	if( Result == 0 )
	{
		//	hCȍԂIԂɂ
		EffectDriver->Quit	= TRUE;

		//	[J[ XbhÎ҂
		for(;;)
		{
			GetExitCodeThread( EffectDriver->hThread, &ExitCode );
			if( ExitCode != STILL_ACTIVE )
			{
				//	Xbh
				CloseHandle( EffectDriver->hThread );
				break;
			}
			//	1ms ҂
			Sleep( 1 );
		}

		//	foCX C^[tFCX
		CloseHandle( EffectDriver->hDevice );

		//	J
		LocalFree( EffectDriver );
	}

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_DeviceID
//----------------------------------------------------------------------------------------------
STDMETHODIMP EffectDriver_DeviceID(
	 IDirectInputEffectDriver *	EffectDriverInterface
	,DWORD						DiVersion
	,DWORD						ExternalID
	,DWORD						Begin
	,DWORD						InternalID
	,LPDIHIDFFINITINFO			DiHIDFFInitInfo )
{
	//	ϐ錾
	HRESULT			Result			= S_OK;
	EFFECT_DRIVER *	EffectDriver	= (PEFFECT_DRIVER)EffectDriverInterface;
	_TCHAR			DevicePath[MAX_PATH];
	DWORD			ThreadID;

#ifdef	UNICODE
	_tcscpy( DevicePath, DiHIDFFInitInfo->pwszDeviceInterface );
#else
	WideCharToMultiByte(
		 CP_ACP
		,0
		,(LPCWSTR)DiHIDFFInitInfo->pwszDeviceInterface
		,-1
		,DevicePath
		,sizeof(_TCHAR) * MAX_PATH
		,NULL
		,NULL );
#endif

	//	foCX C^[tFCXJ
	EffectDriver->hDevice	= CreateFile(
								 DevicePath
								,GENERIC_READ | GENERIC_WRITE
								,FILE_SHARE_READ | FILE_SHARE_WRITE
								,NULL
								,OPEN_EXISTING
								,FILE_ATTRIBUTE_NORMAL
								,NULL );
	if( EffectDriver->hDevice == INVALID_HANDLE_VALUE )
	{
		return( E_FAIL );
	}

	//	[J[ XbhJn
	EffectDriver->hThread	= CreateThread(
								 NULL
								,0
								,EffectProc
								,(LPVOID)EffectDriver
								,0
								,&ThreadID );
	if( EffectDriver->hThread == NULL )
	{
		return( E_FAIL );
	}

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_GetVersions
//----------------------------------------------------------------------------------------------
STDMETHODIMP EffectDriver_GetVersions(
	 IDirectInputEffectDriver *	EffectDriverInterface
	,LPDIDRIVERVERSIONS			DriverVersions )
{
	//	ϐ錾
	HRESULT	Result	= S_OK;

	//	hCo o[W̃TCY`FbN
	if( DriverVersions->dwSize != sizeof( DIDRIVERVERSIONS ) )
	{
		return( E_INVALIDARG );
	}

	//	hCo[ o[WԂ
	DriverVersions->dwFirmwareRevision	= 1;
	DriverVersions->dwHardwareRevision	= 1;
	DriverVersions->dwFFDriverVersion	= 1;

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_Escape
//----------------------------------------------------------------------------------------------
STDMETHODIMP EffectDriver_Escape(
	 IDirectInputEffectDriver *	EffectDriverInterface
	,DWORD						ExternalID
	,DWORD						Effect
	,LPDIEFFESCAPE				DIEffEscape )
{
	//	ϐ錾
	HRESULT	Result	= S_OK;

	//	T|[g
	Result	= E_NOTIMPL;

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_SetGain
//----------------------------------------------------------------------------------------------
STDMETHODIMP EffectDriver_SetGain(
	 IDirectInputEffectDriver *	EffectDriverInterface
	,DWORD						ExternalID
	,DWORD						Gain )
{
	//	ϐ錾
	HRESULT			Result			= S_OK;
	EFFECT_DRIVER *	EffectDriver	= (PEFFECT_DRIVER)EffectDriverInterface;

	//	NeBJ ZNVɓ
	EnterCritical();

	//	QCݒ肷
	if( ( 1 <= Gain ) && ( Gain <= 10000 ) )
	{
		EffectDriver->Gain	= Gain;
	} else {
		EffectDriver->Gain	= max( 1, min( 10000, Gain ) );
		Result				= DI_TRUNCATED;
	}

	//	NeBJ ZNVo
	LeaveCritical();

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_SendForceFeedbackCommand
//----------------------------------------------------------------------------------------------
STDMETHODIMP EffectDriver_SendForceFeedbackCommand(
	 IDirectInputEffectDriver *	EffectDriverInterface
	,DWORD						ExternalID
	,DWORD						Command )
{
	//	ϐ錾
	HRESULT			Result			= S_OK;
	EFFECT_DRIVER *	EffectDriver	= (PEFFECT_DRIVER)EffectDriverInterface;
	PLINK_LIST		EffectList;
	PLINK_LIST		NextEffectList;
	LONG			ContinueTime;

	//	NeBJ ZNVɓ
	EnterCritical();

	//	R}hɂďU蕪
	switch( Command ) {

		case DISFFC_RESET:
			//	SẴGtFNg폜
			EffectList	= EffectDriver->EffectList;
			if( EffectList != NULL )
			{
				do
				{
					//	̃GtFNg XgێĂ
					NextEffectList	= EffectList->Next;

					//	GtFNg폜
					DestroyEffect( &EffectDriver->EffectList, EffectList->Handle );

					//	̃GtFNg Xg擾
					EffectList	= NextEffectList;
				}
				while( EffectList != NULL);
			}
			EffectDriver->Pause	= FALSE;
			break;

		case DISFFC_STOPALL:
			//	SẴGtFNg̍Đ~
			EffectList	= EffectDriver->EffectList;
			if( EffectList != NULL )
			{
				do
				{
					//	GtFNg~
					StopEffect( EffectList->Effect );

					//	̃GtFNg Xg擾
					EffectList	= EffectList->Next;
				}
				while( EffectList != NULL);
			}
			EffectDriver->Pause	= FALSE;
			break;

		case DISFFC_PAUSE:
			//	Đꎞ~
			EffectDriver->PauseTime	= GetTickCount();
			EffectDriver->Pause		= TRUE;
			break;

		case DISFFC_CONTINUE:
			//	ꎞ~
			EffectList	= EffectDriver->EffectList;
			if( EffectList != NULL )
			{
				//	ꎞ~ԂGtFNg̊JnԂx点
				ContinueTime	= GetTickCount() - EffectDriver->PauseTime;
				do
				{
					//	GtFNg̍ĐԂx点
					EffectList->Effect->StartTime	+= ContinueTime;

					//	̃GtFNg Xg擾
					EffectList	= EffectList->Next;
				}
				while( EffectList != NULL);
			}
			EffectDriver->Pause	= FALSE;
			break;

		case DISFFC_SETACTUATORSON:
			//	AN`G[^Lɂ
			EffectDriver->Actuators	= TRUE;
			break;

		case DISFFC_SETACTUATORSOFF:
			//	AN`G[^𖳌ɂ
			EffectDriver->Actuators	= FALSE;
			break;

		default:
			Result	= E_NOTIMPL;
			break;
	}

	//	NeBJ ZNVo
	LeaveCritical();

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_GetForceFeedbackState
//----------------------------------------------------------------------------------------------
STDMETHODIMP EffectDriver_GetForceFeedbackState(
	 IDirectInputEffectDriver *	EffectDriverInterface
	,DWORD						ExternalID
	,LPDIDEVICESTATE			DeviceState )
{
	//	ϐ錾
	HRESULT			Result			= S_OK;
	EFFECT_DRIVER *	EffectDriver	= (PEFFECT_DRIVER)EffectDriverInterface;

	//	foCXԂ̃TCY`FbN
	if( DeviceState->dwSize != sizeof( DIDEVICESTATE ) )
	{
		return( E_INVALIDARG );
	}

	//	NeBJ ZNVɓ
	EnterCritical();

	//	GtFNg hCo[̏ԂԂ
	DeviceState->dwState	= 0;
	//	GtFNg Xg͋󂩂ǂ
	if( EffectDriver->EffectList == NULL )
	{
		DeviceState->dwState	|= DIGFFS_EMPTY;
	}
	//	GtFNg͒~ǂ
	if( (EffectDriver->Playing == FALSE )
	&&	(EffectDriver->Pause   == FALSE ) )
	{
		DeviceState->dwState	|= DIGFFS_STOPPED;
	}
	//	GtFNg͈ꎞ~ǂ
	if( EffectDriver->Pause == TRUE )
	{
		DeviceState->dwState	|= DIGFFS_PAUSED;
	}
	//	AN`G[^͍쓮Ă邩ǂ
	if( EffectDriver->Actuators == TRUE )
	{
		DeviceState->dwState	|= DIGFFS_ACTUATORSON;
	} else {
		DeviceState->dwState	|= DIGFFS_ACTUATORSOFF;
	}
	//	d̓IŒ
	DeviceState->dwState	|= DIGFFS_POWERON;
	//	Z[teB XCb`̓ItŒ
	DeviceState->dwState	|= DIGFFS_SAFETYSWITCHOFF;
	//	[U[ XCb`̓IŒ
	DeviceState->dwState	|= DIGFFS_USERFFSWITCHON;

	//	[h\͖Œ
	DeviceState->dwLoad		= 0;

	//	NeBJ ZNVo
	LeaveCritical();

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_DownloadEffect
//----------------------------------------------------------------------------------------------
STDMETHODIMP EffectDriver_DownloadEffect(
	 IDirectInputEffectDriver *	EffectDriverInterface
	,DWORD						ExternalID
	,DWORD						EffectID
	,LPDWORD					EffectHandle
	,LPCDIEFFECT				DiEffect
	,DWORD						Flags )
{
	//	ϐ錾
	HRESULT			Result			= S_OK;
	EFFECT_DRIVER *	EffectDriver	= (PEFFECT_DRIVER)EffectDriverInterface;
	PEFFECT			Effect;

	//	DIEP_NODOWNLOAD ̏ꍇ͖
	if( Flags & DIEP_NODOWNLOAD )
	{
		return( Result );
	}

	//	NeBJ ZNVɓ
	EnterCritical();

	//	VKGtFNg쐬
	if( *EffectHandle == 0 )
	{
		*EffectHandle	= ( EffectDriver->EffectCount ++ );
		Effect			= LocalAlloc( LPTR | LMEM_ZEROINIT, sizeof( EFFECT ) );
		AddLink( &EffectDriver->EffectList, *EffectHandle, Effect );
	} else {
		Effect			= SearchLink( &EffectDriver->EffectList, *EffectHandle );
	}

	//	GtFNghcݒ肷
	Effect->EffectID	= EffectID;

	//	Flags ݒ肷
	Effect->DiEffect.dwFlags			= DiEffect->dwFlags;

	//	DIEP_DURATION ݒ肷
	if( Flags & DIEP_DURATION )
	{
		Effect->DiEffect.dwDuration			= DiEffect->dwDuration;
	}

	//	DIEP_SAMPLEPERIOD ݒ肷
	if( Flags & DIEP_SAMPLEPERIOD )
	{
		Effect->DiEffect.dwSamplePeriod		= DiEffect->dwSamplePeriod;
	}

	//	DIEP_GAIN ݒ肷
	if( Flags & DIEP_GAIN )
	{
		Effect->DiEffect.dwGain				= DiEffect->dwGain;
	}

	//	DIEP_TRIGGERBUTTON ݒ肷
	if( Flags & DIEP_TRIGGERBUTTON )
	{
		Effect->DiEffect.dwTriggerButton	= DiEffect->dwTriggerButton;
	}

	//	DIEP_TRIGGERREPEATINTERVAL ݒ肷
	if( Flags & DIEP_TRIGGERREPEATINTERVAL )
	{
		Effect->DiEffect.dwTriggerRepeatInterval
											= DiEffect->dwTriggerRepeatInterval;
	}

	//	DIEP_AXES ݒ肷
	if( Flags & DIEP_AXES )
	{
		Effect->DiEffect.cAxes				= DiEffect->cAxes;
		if( Effect->Axes == NULL )
		{
			Effect->Axes						= (DWORD *)LocalAlloc(
													 LPTR | LMEM_ZEROINIT
													,sizeof( DWORD ) * DiEffect->cAxes );
		}
		CopyMemory(
			 Effect->Axes
			,DiEffect->rgdwAxes
			,sizeof( DWORD ) * DiEffect->cAxes );
		Effect->DiEffect.rgdwAxes			= Effect->Axes;
	}

	//	DIEP_DIRECTION ݒ肷
	if( Flags & DIEP_DIRECTION )
	{
		Effect->DiEffect.cAxes				= DiEffect->cAxes;
		if( Effect->Direction == NULL )
		{
			Effect->Direction					= (LONG *)LocalAlloc(
													 LPTR | LMEM_ZEROINIT
													,sizeof( LONG ) * DiEffect->cAxes );
		}
		CopyMemory(
			 Effect->Direction
			,DiEffect->rglDirection
			,sizeof( LONG ) * DiEffect->cAxes );
		Effect->DiEffect.rglDirection		= Effect->Direction;
	}

	//	DIEP_ENVELOPE ݒ肷
	if( ( Flags & DIEP_ENVELOPE )
	&&	( DiEffect->lpEnvelope != NULL ) )
	{
		CopyMemory( &Effect->DiEnvelope, DiEffect->lpEnvelope, sizeof( DIENVELOPE ) );
		//	A^bNԂƃtF[hԂԂꍇ͒l␳
		if( Effect->DiEffect.dwDuration - Effect->DiEnvelope.dwFadeTime
				< Effect->DiEnvelope.dwAttackTime )
		{
			Effect->DiEnvelope.dwFadeTime		= Effect->DiEnvelope.dwAttackTime;
		}
		Effect->DiEffect.lpEnvelope			= &Effect->DiEnvelope;
	}

	//	TypeSpecificParams ݒ肷
	Effect->DiEffect.cbTypeSpecificParams		= DiEffect->cbTypeSpecificParams;

	//	DIEP_TYPESPECIFICPARAMS ݒ肷
	if( Flags & DIEP_TYPESPECIFICPARAMS )
	{
		switch( EffectID )
		{
			case SPRING:
			case DAMPER:
			case INERTIA:
			case FRICTION:
				if( Effect->DiCondition == NULL )
				{
					Effect->DiCondition					= (DICONDITION *)LocalAlloc(
															 LPTR | LMEM_ZEROINIT
															,DiEffect->cbTypeSpecificParams );
				}
				CopyMemory(
					 Effect->DiCondition
					,DiEffect->lpvTypeSpecificParams
					,DiEffect->cbTypeSpecificParams );
				Effect->DiEffect.lpvTypeSpecificParams	= Effect->DiCondition;
				break;

			case CONSTANT_FORCE:
				CopyMemory(
					 &Effect->DiConstantForce
					,DiEffect->lpvTypeSpecificParams
					,DiEffect->cbTypeSpecificParams );
				Effect->DiEffect.lpvTypeSpecificParams	= &Effect->DiConstantForce;
				break;

			case CUSTOM_FORCE:
				if( Effect->DiCustomForce == NULL )
				{
					Effect->DiCustomForce				= (DICUSTOMFORCE *)LocalAlloc(
															 LPTR | LMEM_ZEROINIT
															,DiEffect->cbTypeSpecificParams );
				}
				CopyMemory(
					 Effect->DiCustomForce
					,DiEffect->lpvTypeSpecificParams
					,DiEffect->cbTypeSpecificParams );
				Effect->DiEffect.lpvTypeSpecificParams	= Effect->DiCustomForce;
				break;

			case SQUARE:
			case SINE:
			case TRIANGLE:
			case SAWTOOTH_UP:
			case SAWTOOTH_DOWN:
				CopyMemory(
					 &Effect->DiPeriodic
					,DiEffect->lpvTypeSpecificParams
					,DiEffect->cbTypeSpecificParams );
				Effect->DiEffect.lpvTypeSpecificParams	= &Effect->DiPeriodic;
				break;

			case RAMP_FORCE:
				CopyMemory(
					 &Effect->DiRampforce
					,DiEffect->lpvTypeSpecificParams
					,DiEffect->cbTypeSpecificParams );
				Effect->DiEffect.lpvTypeSpecificParams	= &Effect->DiRampforce;
				break;
		}
	}

	//	DIEP_STARTDELAY ݒ肷
	if( Flags & DIEP_STARTDELAY )
	{
		Effect->DiEffect.dwStartDelay		= DiEffect->dwStartDelay;
	}

	//	DIEP_START w肳Ă΃GtFNgĐ
	if( Flags & DIEP_START )
	{
		StartEffect( Effect, 1 );
	}

	//	DIEP_NORESTART ͖
	if( Flags & DIEP_NORESTART )
	{
		;
	}

	//	NeBJ ZNVo
	LeaveCritical();

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_DestroyEffect
//----------------------------------------------------------------------------------------------
STDMETHODIMP EffectDriver_DestroyEffect(
	 IDirectInputEffectDriver *	EffectDriverInterface
	,DWORD						ExternalID
	,DWORD						EffectHandle )
{
	//	ϐ錾
	HRESULT			Result			= S_OK;
	EFFECT_DRIVER *	EffectDriver	= (PEFFECT_DRIVER)EffectDriverInterface;

	//	NeBJ ZNVɓ
	EnterCritical();

	//	GtFNg폜
	DestroyEffect( &EffectDriver->EffectList, EffectHandle );

	//	NeBJ ZNVo
	LeaveCritical();

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	DestroyEffect
//----------------------------------------------------------------------------------------------
STDAPI_(void) DestroyEffect(
	 PLINK_LIST *	EffectList
	,DWORD			EffectHandle )
{
	//	ϐ錾
	PEFFECT	Effect;

	//	nhGtFNg
	Effect	= SearchLink( EffectList, EffectHandle );

	//	GtFNg폜
	if( Effect->Axes != NULL )
	{
		LocalFree( (VOID *)Effect->Axes );
	}
	if( Effect->Direction != NULL )
	{
		LocalFree( (VOID *)Effect->Direction );
	}
	if( Effect->DiCondition != NULL )
	{
		LocalFree( (VOID *)Effect->DiCondition );
	}
	if( Effect->DiCustomForce != NULL )
	{
		LocalFree( (VOID *)Effect->DiCustomForce );
	}
	LocalFree( (VOID *)Effect );
	DelLink( EffectList, EffectHandle );
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_StartEffect
//----------------------------------------------------------------------------------------------
STDMETHODIMP EffectDriver_StartEffect(
	 IDirectInputEffectDriver *	EffectDriverInterface
	,DWORD						ExternalID
	,DWORD						EffectHandle
	,DWORD						Mode
	,DWORD						Count )
{
	//	ϐ錾
	HRESULT			Result			= S_OK;
	EFFECT_DRIVER *	EffectDriver	= (PEFFECT_DRIVER)EffectDriverInterface;
	PEFFECT			Effect;
	PLINK_LIST		EffectList;

	//	NeBJ ZNVɓ
	EnterCritical();

	//	nhGtFNg
	Effect	= SearchLink( &EffectDriver->EffectList, EffectHandle );

	//	GtFNgĐ
	StartEffect( Effect, Count );

	//	DIES_SOLO w肳Ăꍇ͑̃GtFNg~
	if( Mode & DIES_SOLO )
	{
		EffectList	= EffectDriver->EffectList;
		do
		{
			//	GtFNg~
			if( EffectList->Handle != EffectHandle )
			{
				//	GtFNg~
				StopEffect( EffectList->Effect );
			}
			//	̃GtFNg Xg擾
			EffectList	= EffectList->Next;
		}
		while( EffectList != NULL);
	}

	//	NeBJ ZNVo
	LeaveCritical();

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	StartEffect
//----------------------------------------------------------------------------------------------
STDAPI_(void) StartEffect(
	 PEFFECT	Effect
	,DWORD		Count )
{
	//	GtFNgĐԂɂ
	Effect->Status		= DIEGES_PLAYING;
	Effect->Count		= Count;
	Effect->StartTime	= GetTickCount();
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_StopEffect
//----------------------------------------------------------------------------------------------
STDMETHODIMP EffectDriver_StopEffect(
	 IDirectInputEffectDriver *	EffectDriverInterface
	,DWORD						ExternalID
	,DWORD						EffectHandle )
{
	//	ϐ錾
	HRESULT			Result			= S_OK;
	EFFECT_DRIVER *	EffectDriver	= (PEFFECT_DRIVER)EffectDriverInterface;
	PEFFECT			Effect;

	//	NeBJ ZNVɓ
	EnterCritical();

	//	nhGtFNg
	Effect	= SearchLink( &EffectDriver->EffectList, EffectHandle );

	//	GtFNg~
	StopEffect( Effect );

	//	NeBJ ZNVo
	LeaveCritical();

	return( Result );
}

//----------------------------------------------------------------------------------------------
//	StopEffect
//----------------------------------------------------------------------------------------------
STDAPI_(void) StopEffect( PEFFECT Effect )
{
	//	GtFNg~
	Effect->Status	= 0;
}

//----------------------------------------------------------------------------------------------
//	EffectDriver_GetEffectStatus
//----------------------------------------------------------------------------------------------
STDMETHODIMP EffectDriver_GetEffectStatus(
	 IDirectInputEffectDriver *	EffectDriverInterface
	,DWORD						ExternalID
	,DWORD						EffectHandle
	,LPDWORD					Status )
{
	//	ϐ錾
	HRESULT			Result			= S_OK;
	EFFECT_DRIVER *	EffectDriver	= (PEFFECT_DRIVER)EffectDriverInterface;
	PEFFECT			Effect;

	//	NeBJ ZNVɓ
	EnterCritical();

	//	nhGtFNg
	Effect	= SearchLink( &EffectDriver->EffectList, EffectHandle );

	//	GtFNg̏ԂԂ
	*Status	= Effect->Status;

	//	NeBJ ZNVo
	LeaveCritical();

	return( Result );
}
