
//	Force Feedback Driver for Microsoft Xbox 360 Controller

#include	"X360CFF.h"

//----------------------------------------------------------------------------------------------
//	EffectProc
//----------------------------------------------------------------------------------------------
STDAPI_(DWORD) WINAPI EffectProc( LPVOID EffectDriverInterface )
{
	//	ϐ錾
	EFFECT_DRIVER *	EffectDriver	= (PEFFECT_DRIVER)EffectDriverInterface;
	PLINK_LIST		WorkList;
	PEFFECT			Effect;
	ULONG			CurrentTime;
	LONG			LeftLevel;
	LONG			RightLevel;
	BYTE			Output[4];
	DWORD			ReqLen;

	while( EffectDriver->Quit == FALSE )
	{
		//	NeBJ ZNVɓ
		EnterCritical();

		//	[vɍXVl
		EffectDriver->Playing	= FALSE;
		LeftLevel				= 0;
		RightLevel				= 0;

		WorkList	= EffectDriver->EffectList;
		if(	( WorkList != NULL )
		&&	( EffectDriver->Actuators == TRUE ) )
		{
			//	ݎ擾
			CurrentTime		= GetTickCount();
			do
			{
				//	GtFNg XgGtFNg擾
				Effect	= WorkList->Effect;

				//	GtFNgvZ
				CalcEffect( Effect, CurrentTime, &LeftLevel, &RightLevel );

				//	̃GtFNg Xg擾
				WorkList	= WorkList->Next;
			}
			while( WorkList != NULL);
		}

		//	NeBJ ZNVo
		LeaveCritical();

		//	foCXփtH[Xo
		Output[0]	= 0;
		Output[1]	= (BYTE)min( 255, LeftLevel * EffectDriver->Gain / 10000 );
		Output[2]	= (BYTE)min( 255, RightLevel * EffectDriver->Gain / 10000 );
		Output[3]	= 0;
		WriteFile( EffectDriver->hDevice, Output, sizeof( Output ), &ReqLen, NULL );

		//	2ms ҋ@
		Sleep( 2 );
	}

	//	foCXփtH[Xo
	Output[0]	= 0;
	Output[1]	= 0;
	Output[2]	= 0;
	Output[3]	= 0;
	WriteFile( EffectDriver->hDevice, Output, sizeof( Output ), &ReqLen, NULL );

	return( 0 );
}

//----------------------------------------------------------------------------------------------
//	CalcEffect
//----------------------------------------------------------------------------------------------
STDAPI_(void) CalcEffect(
	 EFFECT *	Effect
	,ULONG		CurrentTime
	,LONG *		LeftLevel
	,LONG *		RightLevel )
{
	//	ϐ錾
	BOOL	Result	= TRUE;
	ULONG	Duration;
	ULONG	StartTime;
	ULONG	EndTime;
	ULONG	ElapsedTime;
	LONG	NormalLevel;
	LONG	NormalRate;
	LONG	AttackLevel;
	LONG	FadeLevel;
	LONG	WorkLeftLevel;
	LONG	WorkRightLevel;

	//	GtFNg̊JnAIAĐԂZo
	Duration	= Effect->DiEffect.dwDuration / 1000;
	StartTime	= Effect->StartTime + ( Effect->DiEffect.dwStartDelay / 1000 );
	if( Effect->Count != -1 )
	{
		EndTime		= StartTime + Duration * Effect->Count;
	} else {
		EndTime		= 0xFFFFFFFF;
	}
	ElapsedTime	= CurrentTime - StartTime;

	//	GtFNg͍ĐH
	if(	( Effect->Status == DIEGES_PLAYING )
	&&	( StartTime <= CurrentTime )
	&&	( CurrentTime <= EndTime ) )
	{
		//	Gx[vvZ
		CalcEnvelope(
			 &Effect->DiEnvelope
			,Duration
			,ElapsedTime % Duration
			,&NormalRate
			,&AttackLevel
			,&FadeLevel );

		//	tH[XvZ
		CalcForce(
			 Effect
			,Duration
			,ElapsedTime % Duration
			,NormalRate
			,AttackLevel
			,FadeLevel
			,&NormalLevel );

/*
		//	vZ
		CalcDirection(
			 Effect
			,NormalLevel
			,&WorkLeftLevel
			,&WorkRightLevel );

		//	tH[X̐𒲐
		WorkLeftLevel	= (WorkLeftLevel > 0) ? WorkLeftLevel : -WorkLeftLevel;
		WorkRightLevel	= (WorkRightLevel > 0) ? WorkRightLevel : -WorkRightLevel;
*/

		//	tH[X̐𒲐
		WorkLeftLevel	= (NormalLevel > 0) ? NormalLevel : -NormalLevel;
		WorkRightLevel	= (NormalLevel > 0) ? NormalLevel : -NormalLevel;
		//	tH[X̏𒲐
		WorkLeftLevel	= min( 255, WorkLeftLevel * 10000 / 390625 );
		WorkRightLevel	= min( 255, WorkRightLevel * 10000 / 390625 );

		//	tH[XZ
		*LeftLevel	= *LeftLevel + WorkLeftLevel;
		*RightLevel	= *RightLevel + WorkRightLevel;
	}
}

//----------------------------------------------------------------------------------------------
//	CalcEnvelope
//----------------------------------------------------------------------------------------------
STDAPI_(void) CalcEnvelope(
	 DIENVELOPE *	Envelope
	,ULONG			Duration
	,ULONG			ElapsedTime
	,LONG *			NormalRate
	,LONG *			AttackLevel
	,LONG *			FadeLevel )
{
	//	ϐ錾
	ULONG	AttackTime		= Envelope->dwAttackTime / 1000;
	ULONG	FadeTime		= Envelope->dwFadeTime / 1000;
	ULONG	FadeStartTime	= Duration - FadeTime;
	LONG	AttackRate;
	LONG	FadeRate;

	//	Gx[vvZ
	if( Envelope != NULL )
	{
		//	A^bN̊Zo
		if( ElapsedTime < AttackTime )
		{
			AttackRate	= ( AttackTime - ElapsedTime ) * 100 / AttackTime;
		} else {
			AttackRate	= 0;
		}

		//	tF[h̊Zo
		if( FadeStartTime < ElapsedTime )
		{
			FadeRate	= ( ElapsedTime - FadeStartTime ) * 100 / FadeTime;
		} else {
			FadeRate	= 0;
		}

		*NormalRate		= 100 - AttackRate - FadeRate;
		*AttackLevel	= Envelope->dwAttackLevel * AttackRate;
		*FadeLevel		= Envelope->dwFadeLevel * FadeRate;
	} else {
		*NormalRate		= 100;
		*AttackLevel	= 0;
		*FadeLevel		= 0;
	}
}

//----------------------------------------------------------------------------------------------
//	CalcForce
//----------------------------------------------------------------------------------------------
STDAPI_(void) CalcForce(
	 EFFECT *	Effect
	,ULONG		Duration
	,ULONG		ElapsedTime
	,LONG		NormalRate
	,LONG		AttackLevel
	,LONG		FadeLevel
	,LONG *		NormalLevel )
{
	//	ϐ錾
	ULONG	LeftTime	= Duration - ElapsedTime;
	LONG	Rate;
	LONG	R;
	LONG	Period;
	LONG	Magnitude;

	//	GtFNg̎ނɂďU蕪
	switch( Effect->EffectID )
	{
		//	
		case SPRING:
		case DAMPER:
		case INERTIA:
		case FRICTION:
			Magnitude	= 0;
			break;

		//	RX^g tH[X
		case CONSTANT_FORCE:
			//	Gx[vl}Oj`[h߂
			Magnitude	= Effect->DiConstantForce.lMagnitude;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			break;

		//	JX^ tH[X
		case CUSTOM_FORCE:
			Magnitude	= 0;
			break;

		//	IGtFNg
		case SQUARE:
			//	P̎ԁi~bjƌoߎԂ 0 ` 359 x̂ǂ߂
			Period	= max( 1, ( Effect->DiPeriodic.dwPeriod / 1000 ) );
			R		= ( ElapsedTime % Period ) * 360 / Period;
			//	tF[YvZ
			R		= ( R + ( Effect->DiPeriodic.dwPhase / 100 ) ) % 360;
			//	Gx[vl}Oj`[h߂
			Magnitude	= Effect->DiPeriodic.dwMagnitude;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			//	`l}Oj`[h߂
			if( 180 <= R )
			{
				Magnitude	= Magnitude * -1;
			}
			//	ItZbgvZ
			Magnitude	= Magnitude + Effect->DiPeriodic.lOffset;
			break;

		case SINE:
			//	P̎ԁi~bjƌoߎԂ 0 ` 359 x̂ǂ߂
			Period	= max( 1, ( Effect->DiPeriodic.dwPeriod / 1000 ) );
			R		= ( ElapsedTime % Period ) * 360 / Period;
			//	tF[YvZ
			R		= ( R + ( Effect->DiPeriodic.dwPhase / 100 ) ) % 360;
			//	Gx[vl}Oj`[h߂
			Magnitude	= Effect->DiPeriodic.dwMagnitude;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			//	gl}Oj`[h߂
			Magnitude	= (LONG)( Magnitude * sin( R * 3.1415 / 180.0 ) );
			//	ItZbgvZ
			Magnitude	= Magnitude + Effect->DiPeriodic.lOffset;
			break;

		case TRIANGLE:
			//	P̎ԁi~bjƌoߎԂ 0 ` 359 x̂ǂ߂
			Period	= max( 1, ( Effect->DiPeriodic.dwPeriod / 1000 ) );
			R		= ( ElapsedTime % Period ) * 360 / Period;
			//	tF[YvZ
			R		= ( R + ( Effect->DiPeriodic.dwPhase / 100 ) ) % 360;
			//	Gx[vl}Oj`[h߂
			Magnitude	= Effect->DiPeriodic.dwMagnitude;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			//	Opgl}Oj`[h߂
			if( 0 <= R && R < 90 )
			{
				Magnitude	= -Magnitude * ( 90 - R ) / 90;
			}
			if( 90 <= R && R < 180 )
			{
				Magnitude	= Magnitude * ( R - 90 ) / 90;
			}
			if( 180 <= R && R < 270 )
			{
				Magnitude	= Magnitude * ( 90 - ( R - 180 ) ) / 90;
			}
			if( 270 <= R && R < 360 )
			{
				Magnitude	= -Magnitude * ( R - 270 ) / 90;
			}
			//	ItZbgvZ
			Magnitude	= Magnitude + Effect->DiPeriodic.lOffset;
			break;

		case SAWTOOTH_UP:
			//	P̎ԁi~bjƌoߎԂ 0 ` 359 x̂ǂ߂
			Period	= max( 1, ( Effect->DiPeriodic.dwPeriod / 1000 ) );
			R		= ( ElapsedTime % Period ) * 360 / Period;
			//	tF[YvZ
			R		= ( R + ( Effect->DiPeriodic.dwPhase / 100 ) ) % 360;
			//	Gx[vl}Oj`[h߂
			Magnitude	= Effect->DiPeriodic.dwMagnitude;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			//	Abvl}Oj`[h߂
			if( 0 <= R && R < 180 )
			{
				Magnitude	= -Magnitude * ( 180 - R ) / 180;
			}
			if( 180 <= R && R < 360 )
			{
				Magnitude	= Magnitude * ( R - 180 ) / 180;
			}
			//	ItZbgvZ
			Magnitude	= Magnitude + Effect->DiPeriodic.lOffset;
			break;

		case SAWTOOTH_DOWN:
			//	P̎ԁi~bjƌoߎԂ 0 ` 359 x̂ǂ߂
			Period	= max( 1, ( Effect->DiPeriodic.dwPeriod / 1000 ) );
			R		= ( ElapsedTime % Period ) * 360 / Period;
			//	tF[YvZ
			R		= ( R + ( Effect->DiPeriodic.dwPhase / 100 ) ) % 360;
			//	Gx[vl}Oj`[h߂
			Magnitude	= Effect->DiPeriodic.dwMagnitude;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			//	_El}Oj`[h߂
			if( 0 <= R && R < 180 )
			{
				Magnitude	= Magnitude * ( 180 - R ) / 180;
			}
			if( 180 <= R && R < 360 )
			{
				Magnitude	= -Magnitude * ( R - 180 ) / 180;
			}
			//	ItZbgvZ
			Magnitude	= Magnitude + Effect->DiPeriodic.lOffset;
			break;

		//	X΃tH[X
		case RAMP_FORCE:
			//	n_AI_̊Zo
			Rate		= LeftTime * 100 / ( Effect->DiEffect.dwDuration / 1000 );
			//	Gx[vl}Oj`[h߂
			Magnitude	= ( Effect->DiRampforce.lStart * Rate
							+ Effect->DiRampforce.lEnd * ( 100 - Rate ) ) / 100;
			Magnitude	= ( Magnitude * NormalRate + AttackLevel + FadeLevel ) / 100;
			break;
	}

	//	QCltH[XԂ
	*NormalLevel	= Magnitude * (LONG)Effect->DiEffect.dwGain / 10000;
}

/*
//----------------------------------------------------------------------------------------------
//	CalcDirection
//----------------------------------------------------------------------------------------------
STDAPI_(void) CalcDirection(
	 EFFECT *	Effect
	,LONG		NormalLevel
	,LONG *		LeftLevel
	,LONG *		RightLevel )
{
	//	ϐ錾
	LONG	Rate	= 0;
	LONG	R		= 0;

	if( 0 < Effect->DiEffect.cAxes )
	{
		//	W
		if( Effect->DiEffect.dwFlags & DIEFF_CARTESIAN )
		{
			Rate	= Effect->Direction[0] / 256;
		}
		//	ɍW
		if( Effect->DiEffect.dwFlags & DIEFF_POLAR )
		{
			R		= Effect->Direction[0] / 100 - 90;
			Rate	= (LONG)( cos( R * 3.1415 / 180.0 ) * 256 );
		}
		//	W
		if( Effect->DiEffect.dwFlags & DIEFF_SPHERICAL )
		{
			R		= Effect->Direction[0] / 100 - 90;
			Rate	= (LONG)( cos( R * 3.1415 / 180.0 ) * 256 );
		}
	}

	if( 0 < Rate )
	{
		*LeftLevel	= NormalLevel * ( 256 - Rate ) / 256;
		*RightLevel	= NormalLevel;
	}
	if( Rate < 0 )
	{
		*LeftLevel	= NormalLevel;
		*RightLevel	= NormalLevel * ( 256 + Rate ) / 256;
	}
	if( Rate == 0 )
	{
		*LeftLevel	= NormalLevel;
		*RightLevel	= NormalLevel;
	}
}
*/
