
//	Force Feedback Driver for Microsoft Xbox 360 Controller

#define	UNICODE
#define	_UNICODE

#include	<windows.h>
#include	<tchar.h>
#include	<math.h>
#include	<objbase.h>
#include	<initguid.h>
#include	<dinput.h>
#include	<mmsystem.h>
#include	<dinputd.h>

//----------------------------------------------------------------------------------------------
//	萔錾
//----------------------------------------------------------------------------------------------

#define	CONSTANT_FORCE	0x00
#define	RAMP_FORCE		0x01
#define	SQUARE			0x02
#define	SINE			0x03
#define	TRIANGLE		0x04
#define	SAWTOOTH_UP		0x05
#define	SAWTOOTH_DOWN	0x06
#define	SPRING			0x07
#define	DAMPER			0x08
#define	INERTIA			0x09
#define	FRICTION		0x0a
#define	CUSTOM_FORCE	0x0b

//----------------------------------------------------------------------------------------------
//	\̐錾
//----------------------------------------------------------------------------------------------

typedef struct _EFFECT
{
	DWORD			EffectID;
	DIEFFECT		DiEffect;

	DWORD *			Axes;
	LONG *			Direction;

	DIENVELOPE		DiEnvelope;
	DICONSTANTFORCE	DiConstantForce;
	DIPERIODIC		DiPeriodic;
	DIRAMPFORCE		DiRampforce;

	DICONDITION *	DiCondition;
	DICUSTOMFORCE *	DiCustomForce;

	DWORD			Status;
	DWORD			Count;
	DWORD			StartTime;
} EFFECT, *PEFFECT;

typedef struct _LINK_LIST
{
	VOID *		Prev;
	DWORD		Handle;
	EFFECT *	Effect;
	VOID *		Next;
} LINK_LIST, *PLINK_LIST;

typedef struct _EFFECT_DRIVER
{
	IDirectInputEffectDriver	Interface;
	LONG						ReferenceCount;

	PLINK_LIST					EffectList;
	DWORD						EffectCount;

	DWORD						Gain;
	BOOL						Actuators;

	BOOL						Playing;
	BOOL						Pause;
	LONG						PauseTime;

	HANDLE						hDevice;

	HANDLE						hThread;
	BOOL						Quit;
} EFFECT_DRIVER, *PEFFECT_DRIVER;

//----------------------------------------------------------------------------------------------
//	O[oϐ錾
//----------------------------------------------------------------------------------------------

//	Common.c
extern	LONG				ReferenceCount;
extern	CRITICAL_SECTION	CriticalSection;

//----------------------------------------------------------------------------------------------
//	֐錾
//----------------------------------------------------------------------------------------------

//	X360CFF.c
STDAPI_(BOOL) DllMain( HINSTANCE, DWORD, LPVOID );
STDAPI_(void) ProcessAttach( HINSTANCE );
STDAPI_(void) ProcessDetach();

//	Common.c
STDAPI_(LONG) IncrementReference();
STDAPI_(LONG) DecrementReference();
STDAPI_(void) EnterCritical();
STDAPI_(void) LeaveCritical();
STDAPI_(void) AddLink( PLINK_LIST *, DWORD, PEFFECT );
STDAPI_(PEFFECT) SearchLink( PLINK_LIST *, DWORD );
STDAPI_(void) DelLink( PLINK_LIST *, DWORD );

//	COM.c
STDAPI DllGetClassObject( REFCLSID, REFIID, LPVOID * );
STDAPI DllCanUnloadNow();

//	ClassFactory.c
STDMETHODIMP ClassFactory_New( REFIID, LPVOID * );
STDMETHODIMP ClassFactory_QueryInterface( IClassFactory *, REFIID, LPVOID * );
STDMETHODIMP_(ULONG) ClassFactory_AddRef( IClassFactory * );
STDMETHODIMP_(ULONG) ClassFactory_Release( IClassFactory * );
STDMETHODIMP ClassFactory_CreateInstance( IClassFactory *, IUnknown *, REFIID, LPVOID * );
STDMETHODIMP ClassFactory_LockServer( IClassFactory *, BOOL );

//	EffectDriver.c
STDMETHODIMP EffectDriver_New( REFIID, LPVOID * );
STDMETHODIMP EffectDriver_QueryInterface( IDirectInputEffectDriver *, REFIID, LPVOID * );
STDMETHODIMP_(ULONG) EffectDriver_AddRef( IDirectInputEffectDriver * );
STDMETHODIMP_(ULONG) EffectDriver_Release( IDirectInputEffectDriver * );
STDMETHODIMP EffectDriver_DeviceID( IDirectInputEffectDriver *, DWORD, DWORD, DWORD, DWORD, LPDIHIDFFINITINFO );
STDMETHODIMP EffectDriver_GetVersions( IDirectInputEffectDriver *, LPDIDRIVERVERSIONS );
STDMETHODIMP EffectDriver_Escape( IDirectInputEffectDriver *, DWORD, DWORD, LPDIEFFESCAPE );
STDMETHODIMP EffectDriver_SetGain( IDirectInputEffectDriver *, DWORD, DWORD );
STDMETHODIMP EffectDriver_SendForceFeedbackCommand( IDirectInputEffectDriver *, DWORD, DWORD );
STDMETHODIMP EffectDriver_GetForceFeedbackState( IDirectInputEffectDriver *, DWORD, LPDIDEVICESTATE );
STDMETHODIMP EffectDriver_DownloadEffect( IDirectInputEffectDriver *, DWORD, DWORD, LPDWORD, LPCDIEFFECT, DWORD );
STDMETHODIMP EffectDriver_DestroyEffect( IDirectInputEffectDriver *, DWORD, DWORD );
STDAPI_(void) DestroyEffect( PLINK_LIST *, DWORD );
STDMETHODIMP EffectDriver_StartEffect( IDirectInputEffectDriver *, DWORD, DWORD, DWORD, DWORD );
STDAPI_(void) StartEffect( PEFFECT, DWORD );
STDMETHODIMP EffectDriver_StopEffect( IDirectInputEffectDriver *, DWORD, DWORD );
STDAPI_(void) StopEffect( PEFFECT );
STDMETHODIMP EffectDriver_GetEffectStatus( IDirectInputEffectDriver *, DWORD, DWORD, LPDWORD );

//	EffectProc.c
STDAPI_(DWORD) WINAPI EffectProc( LPVOID );
STDAPI_(void) CalcEffect( PEFFECT, ULONG, LONG *, LONG * );
STDAPI_(void) CalcEnvelope( DIENVELOPE *, ULONG, ULONG, LONG *, LONG *, LONG * );
STDAPI_(void) CalcForce( PEFFECT, ULONG, ULONG, LONG, LONG, LONG, LONG * );
/*
STDAPI_(void) CalcDirection( PEFFECT, LONG, LONG *, LONG * );
*/
